/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "Esperance.h"
#include "EsperanceParams.h"

bool* sig_stop = NULL;

static void sigIntHandler(int s) {
  *sig_stop = false;
}

#ifdef CUI
int main(int argc, char** argv) {
  
  Esperance* esp = NULL;

  char out_file[] = "output.mp4";

  if (argc < 2) {
    fprintf(stderr, "usage: Esperance inputfile.m2p\n");
    return EXIT_SUCCESS;
  }

  esp = new Esperance();

  sig_stop = esp->getCtrlBool();
  signal(SIGINT, sigIntHandler);
  
  //ESP:ACTION
  esp->init();  
  
  // allocate memory for each params
  esp->ffmpeg_param = (esp_ffmpeg_params_t*)malloc(sizeof(esp_ffmpeg_params_t));
  esp->x264_param = (esp_x264_params_t*)malloc(sizeof(esp_x264_params_t));
  esp->faac_param = (esp_faac_params_t*)malloc(sizeof(esp_faac_params_t));
  
  //set input file
  //set output file
  esp->ffmpeg_param->inputfile  = argv[1];
  esp->params->outputfileName = out_file;

  //ESP:ACTION
  esp->initDecodeCodec();

  esp->params->fpsMode = 0;
  esp->params->autoDetectProgressive = true;

  esp->params->deinterlace = 1;
  if (esp->getInputWidth() < 600) 
    esp->params->deinterlace = 0;

  esp->params->needResize = true;  
  // set output width height
  if (esp->getAspectX() == 32 && esp->getAspectY() == 27) {
	  esp->params->outputWidth = 854;
	  esp->params->outputHeight = 480;
  } else if (esp->getAspectX() == 8 && esp->getAspectY() == 9) {
  	  esp->params->outputWidth = 720;
	  esp->params->outputHeight = 540;
  } else {
	  esp->params->outputWidth = esp->getInputWidth();
	  esp->params->outputHeight = esp->getInputHeight();
	  esp->params->needResize = false;  
  }
  
  
  // init faac_params
  if (argc > 2) {
    esp->faac_param->bitrate = 160;
  } else {
    esp->faac_param->bitrate = 128;
  }
  esp->faac_param->tns     = 1;
  
  // init x264_params
  if (argc > 2) {
    esp->x264_param->subpel_refine = 7;
    esp->x264_param->frame_reference = 3;
    esp->x264_param->crf = 22;
    esp->x264_param->qp_min = 18;
    esp->x264_param->qp_max = 30;
	esp->x264_param->bframe = 0;
	esp->params->x264_bframe = false;
  } else {
    esp->x264_param->subpel_refine = 5;
    esp->x264_param->frame_reference = 2;
    esp->x264_param->crf = 26;
    esp->x264_param->qp_min = 22;
    esp->x264_param->qp_max = 32;
	esp->x264_param->bframe = 0;
	esp->x264_param->bframe_adaptive = true;
	esp->params->x264_bframe = false;
  }

  esp->x264_param->pass = 0;
  esp->x264_param->e_mode = X264_QMODE_CRF;
  esp->x264_param->qp_step = 4;
  esp->x264_param->me_method = 2;
  esp->x264_param->key_max = 30;
  esp->x264_param->threads = 1;

  esp->x264_param->df_alpha = 0;
  esp->x264_param->df_beta = 0;
  esp->x264_param->nr = 0;
  esp->x264_param->weighted_bipred = true;
  esp->x264_param->bidir_me = true;
  esp->x264_param->fast_pskip = true;
  
  //ESP:ACTION
  esp->initCodecs();

  esp->startEncoding();

  esp->closeCodecs();
  esp->close();

  delete esp;
  
}
#endif

Esperance::Esperance() {
  params = new EsperanceParams();
  
  params->deinterlace = 1;
  params->encodeVideo = true;
  params->encodeAudio = true;

  params->mp4_vSync = false;
  params->mp4_aSync = false;
  
  params->mp4_vDelay = 0;
  
  ctr  = true;
  stop = false;
  susp = false;
  
  videoEncodedFrameCount = 0;
  
  progStDuration = 3003;
  progVideoFrameCount = 0;
  progVideoTotalDuration = 0;
  
}

Esperance::~Esperance() {
  delete params;
}

bool Esperance::init() {

  ffmpeg  = new FFMPEGDecoder();
  x264 = new X264Encoder();
  faac = new FAACEncoder();
  mp4 = new MP4Handler();

  return true;
}

bool Esperance::initDecodeCodec() {
  
  //static int a = 0;
  
  //if (a == 1) return false;

  fprintf(stderr, "esp: init decode codec\n");
  ffmpeg->open(ffmpeg_param);  
  //ffmpeg->open((char*)params->inputfiles[0]esp->ffmpeg_param.inputfile);
  //ffmpeg->open(ffmpeg_param->inputfile);
  //ffmpeg->open((char*)params->inputfiles[0]);
  ffmpeg->init();
  //a = 1;

  return true;
}

bool Esperance::initCodecs() {

  //ffmpeg->open((char*)params->inputfileName);
  //ffmpeg->init();
  
  ffmpeg->setVideoResampleContext(params->outputWidth,
								  params->outputHeight);

  faac_param->samplerate  = ffmpeg->getAudioSampleRate();
  faac_param->numChannels = ffmpeg->getAudioChannels();
		 
  faac->init(faac_param);		 
  faac->open();

  x264_param->width   = params->outputWidth;
  x264_param->height  = params->outputHeight;
  
  x264_param->sar_width  = ffmpeg->getAspectX();
  x264_param->sar_height = ffmpeg->getAspectY(); 
  
  x264_param->fps_den = ffmpeg->getFpsDen();
  x264_param->fps_num = ffmpeg->getFpsNum();
  
  char* x264_state_log =  (char* )malloc(512 + 64);
  char x264_state_log2[512];
  const char fname[] = "/x264_1st_pass_log";
  //getcwd(x264_state_log2,sizeof(x264_state_log2));
  int l = strlen(getenv("HOME"));
  strncpy(x264_state_log2,getenv("HOME"), l);
  x264_state_log2[l] = '\0';
  //strcat(x264_state_log2,getenv("HOME"));
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  strcat(x264_state_log2, fname);
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  sprintf(x264_state_log, "%s", x264_state_log2);
  x264_param->log_file = x264_state_log;
  
  x264->init(x264_param);
  x264->open();

  mp4->init();  
  mp4->open(params->outputfileName);
  mp4->setMP4TimeScale(90000);  
  mp4->setVideoDelayTime(params->mp4_vDelay);

  if (ffmpeg->hasVideo() == true && params->encodeVideo == true) {
    mp4->addH264VideoTrack(x264->getSPSRef(), params->outputWidth, params->outputHeight, 90000);
    mp4->setH264SequenceParameter(x264->getSPSRef(), x264->getSPSLen());
    mp4->setH264PictureParameter (x264->getPPSRef(), x264->getPPSLen());
  }

  if (ffmpeg->hasAudio() == true && params->encodeAudio == true) {
    faac->getDecoderSpecificInfo(mp4->getASCRef(), mp4->getASCLengthRef());
    mp4->addAACAudioTrack(faac->getSampleRate());
  }

  return true;
}

bool Esperance::setParams() {
  return true;
}

bool Esperance::startEncoding() {

  if (x264_param->pass == 2 && x264_param->e_mode == X264_QMODE_CBR ) {//&& params->inputs == 1) {
	  
	encodeX264FirstPass();
	
	if (ctr == false) {
	  return false;
	}
	
	ffmpeg->close();
	delete ffmpeg;
	ffmpeg = new FFMPEGDecoder();
	
	ffmpeg->open(ffmpeg_param);
	ffmpeg->init();
	
	ffmpeg->setVideoResampleContext(params->outputWidth,
									params->outputHeight);
	
	x264_param->pass = 1;
		
	x264->close();
	delete x264;
	
	x264 = new X264Encoder();
	
	x264->init(x264_param);
	x264->open();
	
	videoEncodedFrameCount = 0;
  }

  encodeX264SecondPass();
  	
  for (int i = 1; i < params->inputs && x264_param->pass != 2 && x264_param->e_mode != X264_QMODE_CBR; i++) {
	  
	  ffmpeg->close();
	  delete ffmpeg;
	  ffmpeg = new FFMPEGDecoder();
	  
	  ffmpeg->open((char*)params->inputfiles[i]);
	  ffmpeg->init();
	  
	  ffmpeg->setVideoResampleContext(params->outputWidth,
									  params->outputHeight);
									  
	  params->needResize = true;
	  
	  encodeX264SecondPass();
  }
	 
  flush();  									
  
  return true;
}

bool Esperance::encodeX264FirstPass() {
  int lim = 0;
  
  pic = x264->getX264PictureRef();
  
  //ctr: will be changed by NSThread for INT
  for(; ctr == true; lim++) {
  
	  if (susp == true) {
		  sleep(5);
		  fprintf(stderr, "sys: suspending\n");
		  continue;
	  }
    
    ///// if EOF stop encoding
    if (ffmpeg->isFinished() == true) {
      fprintf(stderr, "dec: EOF done!!\n");
      return true;
    }
    /////
    
    /////
    
    ffmpeg->decode();
    
    /////
    if (ffmpeg->isDoneVideo() != 0 && params->encodeVideo == true) {
      fprintf(stderr, "sys: 1st-pass video processing %d\n", ++videoEncodedFrameCount);
      
      AVPicture* tmpP = NULL;
      
      if (params->deinterlace == 0) {
		  tmpP = ffmpeg->getVideoYUVRef();
      } else {
		  tmpP = ffmpeg->getDeinterlacedVideoYUVRef();
      }
      
      if (params->needResize == true) {
		  copy_frame(ffmpeg->getResampledVideoYUVRef(tmpP), pic);
      } else {
		  copy_frame(tmpP, pic);
      }
      
      pic->i_pts = ffmpeg->getPts();
      //x264->encode(dec->getVideoFrameDuration());
      x264->encode();
    } 
  }
  
  if (ctr == false) {
    fprintf(stderr, "sys: interrupted\n");
  }
  
  return true;
}


bool Esperance::encodeX264SecondPass() {
  int lim = 0;
  int i = 0;
  int k = 0;
  int f = 0;
  
  initialI = true;
  
  pic = x264->getX264PictureRef();
  
  //ctr: will be changed by NSThread for INT
  for(i = 0; ctr == true; lim++) {
  
  	  if (susp == true) {
		  fprintf(stderr, "sys: suspending\n");
		  sleep(5);
		  continue;
	  }
	  
    ///// if EOF stop encoding
	if (ffmpeg->isFinished() == true) {
	  fprintf(stderr, "ffmpeg: EOF done!!\n");
      return true;
    }
    /////

  /////
  if (faac->isNeedBuffer() == false && params->encodeAudio == true) {
  
		  faac->encode();
		  
		  if (faac->getBytesWritten() > 0) {
		    mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
				                     faac->getEncodedBytes(),
				                     faac->getEncodedSamples() / faac->getNumChannels());
	  puts("");
      }
      continue;
    }
    /////

    ffmpeg->decode();

    /////
    if (ffmpeg->isDoneVideo() != 0 && params->encodeVideo == true) {
		fprintf(stderr, "sys: video processing %d\n", ++videoEncodedFrameCount);
		
		AVPicture* tmpP = NULL;
		
		if (params->deinterlace == 1 && ffmpeg->isInterlaced() == true) {
			tmpP = ffmpeg->getDeinterlacedVideoYUVRef();
		} else {
			tmpP = ffmpeg->getVideoYUVRef();		
		}
		
		if (params->needResize == true) {
			copy_frame(ffmpeg->getResampledVideoYUVRef(tmpP), pic);
		} else {
			copy_frame(tmpP, pic);
		}
		
		pic->i_pts = ffmpeg->getPts();
		
		x264->encode();
		
		//if ( x264->isChangedHeader() == true) {
		  // bug is here: don't use this sequence
		  //mp4->setH264SequenceParameter(x264->getSPSRef(), x264->getSPSLen());
		  //mp4->setH264PictureParameter (x264->getPPSRef(), x264->getPPSLen());
			
		//} else 
		
		if ( x264->getBytesWritten() > 0) {
			
			if (params->mp4_vHack == true) {
				f = 1;
			} else {
				f = ((x264->getSliceType() - 1) == 0)? 1 : 0;
			}
			
			if (params->x264_bframe == true) {
				int slice =  x264->getSliceType();
				
				if (slice == 1 || slice == 2) { // I
					handleI();
				} else if (slice == 3) { // P
					handleP();
				} else if (slice == 4 || slice == 5) { // B
					handleB();
				}
				
			} else {
				
				mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
										 x264->getTotalBytesWritten(),
										 this->getVideoFrameDuration(),
										 f);
			}
		}
		
		puts("");
    } else if (ffmpeg->isDoneAudio() != 0 && params->encodeAudio == true) { // AUDIO
	
		fprintf(stderr, "sys: audio processing %d %d\n", ++k, ffmpeg->getDecodedAudioSampleNum());
		
		faac->encode((int32_t *)ffmpeg->getAudioRAWRef(),
					 (int32_t  )ffmpeg->getDecodedAudioSampleNum() / 2);
		
		if (faac->getBytesWritten() > 0) {
			mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
									 faac->getEncodedBytes(),
									 faac->getEncodedSamples() / faac->getNumChannels());
			puts("");
		}
    }
    /////
  }

  if (ctr == false) {
    fprintf(stderr, "sys: interrupted\n");
  }
  
  return true;
}

bool Esperance::flush() {

	int i = 0;
	int f;
	
	do {
	
	    if (ffmpeg->hasVideo() == false) {
			break;
		}

		x264->flush();
		
		if ( x264->getBytesWritten() > 0 ) {
		
			fprintf(stderr, "esp: flushing video\n");
			
			if (params->mp4_vHack == true) {
				f = 1;
			} else {
				f = ((x264->getSliceType() - 1) == 0)? 1 : 0;
			}

			if (params->x264_bframe == true) {
				int slice =  x264->getSliceType();
				
				if (slice == 1 || slice == 2) { // I
					handleI();
				} else if (slice == 3) { // P
					handleP();
				} else if (slice == 4 || slice == 5) { // B
					handleB();
				}
				
			} else {
				mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
										 x264->getTotalBytesWritten(),
										 //ffmpeg->getVideoFrameDuration(),
										 this->getVideoFrameDuration(),
										 f);
			}
			
		}
	} while(x264->getBytesWritten() > 0  && ++i < 10);
	
	if (ffmpeg->hasVideo() == true && params->x264_bframe == true && pSize != 0) {
	
		fprintf(stderr, "esp: flushing last P-Frame\n");
		
	    mp4->writeMP4VideoSample(pBuf,
								 pSize,
								 pDuration,
								 3003,
								 0);
	}
	
	if (params->encodeAudio == true) {
		
		faac->flush();
		
		if (faac->getBytesWritten() > 0) {
		
			fprintf(stderr, "esp: flushing audio\n");
			
		    mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
				                     faac->getEncodedBytes(),
				                     faac->getEncodedSamples() / faac->getNumChannels());
			puts("");
		}
    }
	
	return true;
}

bool Esperance::stopEncoding() {

  ctr = false;
  stop = true;

  return true;
}

bool Esperance::suspendEncoding() {
	if (susp == true) {
		susp = false;
	} else {
		susp = true;
	}
	return true;
}

bool* Esperance::getCtrlBool() {
  return &ctr;
}

bool Esperance::closeCodecs() {

  mp4->close();
  x264->close();
  ffmpeg->close();
  faac->close();

  return true;
}

bool Esperance::close() {

  delete x264;
  delete faac;
  delete ffmpeg;
  delete mp4;

  return true;
}


void Esperance::copy_frame(AVPicture *pin, x264_picture_t  *pout) {

  /* Works very fine. In case 720 x 480 but,,,
  pout->img.i_stride[0] = pin->linesize[0];
  pout->img.i_stride[1] = pin->linesize[1];
  pout->img.i_stride[2] = pin->linesize[2];
  pout->img.i_stride[3] = pin->linesize[3];

  // Y
  memcpy(pout->img.plane[0], pin->data[0],
	 dec->getHeight() * pin->linesize[0]    );
  
  // U
  memcpy(pout->img.plane[1], pin->data[1],
	 dec->getHeight() * pin->linesize[1] / 2);

  // V
  memcpy(pout->img.plane[2], pin->data[2],
	 dec->getHeight() * pin->linesize[2] / 2);
  */

  /* ISSUE: double free
  pout->img.plane[0] = pin->data[0];
  pout->img.plane[1] = pin->data[1];
  pout->img.plane[2] = pin->data[2];
  */

  // slow but stable
  int i, x, y;
  int stride0 = pin->linesize[0];
  int stride1 = pin->linesize[1];
  int stride2 = pin->linesize[2];

  /*
  for (y = 0, i = 0; y < ffmpeg->getHeight(); y++) {
    for (x = 0; x < ffmpeg->getWidth(); x++) {
      pout->img.plane[0][i++] = pin->data[0][stride0 * y + x]; // Y
    }
  }

  for (y = 0, i = 0; y < ffmpeg->getHeight() / 2; y++) {
    for (x = 0; x < ffmpeg->getWidth()/2; x++) {
      pout->img.plane[1][i]   = pin->data[1][stride1 * y + x]; // U
      pout->img.plane[2][i++] = pin->data[2][stride2 * y + x]; // V
    }
  }
  */
  
  for (y = 0, i = 0; y < params->outputHeight; y++) {
    for (x = 0; x < params->outputWidth; x++) {
      pout->img.plane[0][i++] = pin->data[0][stride0 * y + x]; // Y
    }
  }

  for (y = 0, i = 0; y < params->outputHeight / 2; y++) {
    for (x = 0; x < params->outputWidth / 2; x++) {
      pout->img.plane[1][i]   = pin->data[1][stride1 * y + x]; // U
      pout->img.plane[2][i++] = pin->data[2][stride2 * y + x]; // V
    }
  }
  
}			     

int Esperance::getInputWidth() {
  return ffmpeg->getWidth();
}

int Esperance::getOutputWidth() {
  return ffmpeg->getWidth();
}

int Esperance::getInputHeight() {
  return ffmpeg->getHeight();
}

int Esperance::getOutputHeight() {
  return ffmpeg->getHeight();
}

int Esperance::getInputAudioSamplerate() {
  return ffmpeg->getAudioSampleRate();
}

int Esperance::getInputAudioNumChannel() {
  return ffmpeg->getAudioChannels();
}

int Esperance::getFpsDen() {
  return ffmpeg->getFpsDen();
}

int Esperance::getFpsNum() {
  return ffmpeg->getFpsNum();
}

int Esperance::getAspectX() {
  return ffmpeg->getAspectX();
}

int Esperance::getAspectY() {
  return ffmpeg->getAspectY();
}

int Esperance::getVideoEncodedFrameCount() {
  return videoEncodedFrameCount;
}

int64_t Esperance::getDecodedBytes() {
  return ffmpeg->getBytePosition();
}

int64_t Esperance::getInputFileSize() {
 return ffmpeg->getFileSize();
}

void Esperance::handleI() {

  int offset = 0;
  int dul[8];
  
  if (pSize > 0) {
  
    if (bFrames == 0) {
      //offset = ffmpeg->getFrameDuration();
	  //offset = ffmpeg->getVideoFrameDuration();
	  offset = this->getVideoFrameDuration();
    } else {
		//offset = ffmpeg->getFrameDuration() * (bFrames + 1);
	  
	  for (int i = 0; i < bFrames + 1; i++) {
		  //dul[i] = ffmpeg->getVideoFrameDuration();
		  dul[i] = this->getVideoFrameDuration();
		  offset += dul[i];
	  }
	  
    }
    
    mp4->writeMP4VideoSample(
				 pBuf,
			     pSize,
			     pDuration,
			     offset,
			     0);
    pSize = 0;
  }
  
  if (bFrames > 0) {
    int p = 0;
    
    for (int ii = 0; ii < bFrames; ii++) {
      
      mp4->writeMP4VideoSample(bBuf + p,
							   bSize[ii],
							   dul[ii]);
							   //3750,
							   //0);
	  p += bSize[ii];
    }
    bFrames = 0;
    bBufInd = 0;
  }
  
  if (initialI == true) { // avoid initial white I frame
	  mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
							   x264->getTotalBytesWritten(),
							   0,
							   0,
							   1);
	  initialI = false;
  }
  
  mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
						   x264->getTotalBytesWritten(),
						   ffmpeg->getFrameDuration(),
						   ffmpeg->getFrameDuration(),
						   1);
}

/*
 */
void Esperance::handleP() {
  int offset = 0;
  int dul[8];
  
  if (pSize > 0) {
    
    if (bFrames == 0) {
      //offset = ffmpeg->getFrameDuration();
	  //offset = ffmpeg->getVideoFrameDuration();
	  offset = this->getVideoFrameDuration();
    } else {
	
	  for (int i = 0; i < bFrames + 1; i++) {
		  dul[i] = this->getVideoFrameDuration();
		  offset += dul[i];
	  }
	  
	  //offset = ffmpeg->getFrameDuration() * (bFrames + 1);
    }
    
    mp4->writeMP4VideoSample(pBuf,
			     pSize,
			     pDuration,
			     offset,
			     0);
    pSize = 0;
  }
  
  if (bFrames > 0) {
    int p = 0;
    
    for (int ii = 0; ii < bFrames; ii++) {
      /*
      mp4->writeMP4VideoSample(bBuf + p,
			       bSize[ii],
			       bDuration[ii]);
      */
		mp4->writeMP4VideoSample(bBuf + p,
								 bSize[ii],
								 dul[ii+1]);
								 
	  
      p += bSize[ii];
    }
    bFrames = 0;
    bBufInd = 0;
  }
  
  storeP(x264->getNalEncodedBufferRef(),
	 x264->getTotalBytesWritten(),
	 ffmpeg->getFrameDuration());
}

void Esperance::handleB() {
	storeB(x264->getNalEncodedBufferRef(),
		   x264->getTotalBytesWritten(),
		   ffmpeg->getFrameDuration());
}

void Esperance::storeP(u_int8_t* data, int size, int duration) {
  pDuration = duration;
  pSize = size;
  memcpy(pBuf, data, size);
}

void Esperance::storeB(u_int8_t* data, int size, int duration) {
  bSize[bFrames] = size;
  bDuration[bFrames] = duration;
  memcpy(bBuf + bBufInd, data, size);
  
  bBufInd += size;
  bFrames++;
}

int Esperance::getVideoFrameDuration() {
    static int v = 0;
	
	if ( params->fpsMode <= 0 ) {
		if (ffmpeg->isInterlaced() == true) {
			fprintf(stderr, "sys: interlaced\n");
			return 3003;
		} else {
			fprintf(stderr, "sys: progressived\n");
			return 3754;
		}
	} else if ( params->fpsMode == 1 ) {
		return ffmpeg->getVideoFrameDuration();	
	} else if ( params->fpsMode == 2 ) {
		if (ffmpeg->isInterlaced() == true) {
			fprintf(stderr, "sys: interlaced\n");
			return 3003;
		} else {
			fprintf(stderr, "sys: progressived\n");
			if ((v++) & 0x1) {
				return 3753;
			} else {
			    return 3752;
			}
		}
	}
	
	return 3003;
}
