#include <stdlib.h>
#include <inttypes.h>

#include "EsperanceParams.h"

#include "X264Encoder.h"
#include "FFMPEGDecoder.h"
#include "MP4Handler.h"
#include "FAACEncoder.h"

#ifndef _ESPERANCE_H_
#define _ESPERANCE_H_

/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

class Esperance {

 private:
 
  X264Encoder   *x264;
  FAACEncoder   *faac;
  FFMPEGDecoder *ffmpeg;
  MP4Handler    *mp4;
  x264_picture_t *pic;
  
  bool ctr;
  bool stop;
  bool susp;
  
  int videoEncodedFrameCount;
  int progStDuration;
  int progVideoFrameCount;
  int64_t progVideoTotalDuration;
  
  void copy_frame(AVPicture *, x264_picture_t  *);

  void writeVideoFrame(int, int);
  
  int getVideoFrameDuration();
  
  void handleI();
  bool initialI;
  
  void handleP();
  void storeP(u_int8_t*, int, int);
  u_int8_t pBuf[160000];
  int      pSize;
  int      pDuration;

  void handleB();
  void storeB(u_int8_t*, int, int);
  u_int8_t bBuf[200000];
  int bBufInd;
  int bFrames;
  int bSize[5];
  int bDuration[5];
  
 public:
  EsperanceParams *params;
   
  Esperance();
  ~Esperance();
  
  esp_ffmpeg_params_t *ffmpeg_param;
  esp_x264_params_t *x264_param;
  esp_faac_params_t *faac_param;

  bool init();
  
  bool initDecodeCodec();
  bool initCodecs();
  bool initCodecs(const char* filename);

  bool setParams();

  bool startEncoding();
  bool stopEncoding();
  bool suspendEncoding();
  
  bool encodeX264FirstPass();
  bool encodeX264SecondPass();
  bool flush();

  bool closeCodecs(void);
  bool close();
  
  int getInputWidth();
  int getOutputWidth();
  
  int getInputHeight();
  int getOutputHeight();
  
  int getInputAudioSamplerate();
  int getInputAudioNumChannel();
  
  int getFpsDen();
  int getFpsNum();
  
  int getAspectX();
  int getAspectY();
  
  bool* getCtrlBool();
  
  int getVideoEncodedFrameCount();
  
  int64_t getDecodedBytes();
  int64_t getInputFileSize();
};
#endif
