/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#ifndef _FFMPEGDECODER_H_
#define _FFMPEGDECODER_H_

#ifdef _USE_MAKE_
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#include <ffmpeg/avutil.h>
#else
#include "ffmpeg/avcodec.h"
#include "ffmpeg/avformat.h"
#include "ffmpeg/avutil.h"
#endif // _USER_MAKE_

typedef struct {
	char* inputfile;
	int output_style;
	
} esp_ffmpeg_params_t;

class FFMPEGDecoder {
 private:
  esp_ffmpeg_params_t* _params;
  
  AVFormatContext *_avFormatCtx;
  AVFormatParameters *_avFormatPrm;
  AVInputFormat *_avInputFmt;

  AVCodecContext  *_vCodecCtx;
  AVCodec         *_vCodec;
  AVFrame         *_vFrame;
  AVPicture       _vDeinterFrame;
  AVPacket        _vPacket;
  AVPicture       _vResampleFrame;

  AVCodecContext  *_aCodecCtx;
  AVCodec         *_aCodec;  
  int32_t         *_aFrame;
  AVPacket        _aPacket;

  ImgReSampleContext * _vResampleCtx;
  ReSampleContext *  _aResampleCtx;
  
  int16_t         *_aResampleFrame;
  
  AVPacket        _packet;
  
  bool _activated;

  int _videoStream;
  int _audioStream;
  int _remainBytes;
  int _remainVideoBytes;
  int _remainAudioBytes;
  int _decodedVideoBytes;
  int _decodedVideoFrames;
  int _decodedAudioBytes;
  int _decodedAudioFrames;

  int _totalDecodedAudioBytes;
  int _totalDecodedFrames;
  
  int64_t _totalDecodedBytes;

  int _doneVideo;
  int _doneAudio;
  bool _hasVideo;
  bool _hasAudio;

  bool _finished;
  bool _finishedReading;
  bool _finishedVideo;
  bool _finishedAudio;

  uint8_t *_rawVideoBuffer;
  uint8_t *_audioMemory;
  uint8_t *_rawAudioBuffer;
  uint8_t *_resampledAudioBuffer;
  int _width;
  int _height;
  
  int _preVideoFrameDuration;
  bool _changedVideoFrameDuration;
  
  void putVideoDuration(int);
  int  getNextVideoDuration(void);
  int _dRing[32];
  int _dRingSize;
  int _dRingData;
  int _dRingMemory;
  
  int64_t _prePTS;
  
  int _aspectX;
  int _aspectY;
  
  int  _chunkDuration;
  bool _interlaced;
  bool detectProgressive(int);

 public:
  FFMPEGDecoder();
  ~FFMPEGDecoder();
  
  bool open(esp_ffmpeg_params_t* );
  bool open(char *);
  bool init();
  bool init(int64_t );
  bool close();

  bool isFinished();

  bool hasVideo();
  bool hasAudio();

  bool decode();
  bool decodeVideo();
  bool decodeOnlyVideo();
  bool decodeAudio();
  bool decodeOnlyAudio();

  //AVFrame *getVideoYUVRef();
  AVPicture *getVideoYUVRef();
  AVPicture *getDeinterlacedVideoYUVRef();
  AVPicture *getResampledVideoYUVRef();
  AVPicture *getResampledVideoYUVRef(AVPicture *);
  
  uint8_t   *getVideoYRef();
  int        getVideoYSize();
  
  uint8_t   *getVideoURef();
  int        getVideoUSize();
  
  uint8_t   *getVideoVRef();
  int        getVideoVSize();
  
  int64_t getFileSize();
  int64_t getBytePosition();
  int64_t getPts();
  int64_t getDts();
  int getFrameDuration();
  bool didChangeVideoFrameDuration();
  
  int32_t *getAudioRAWRef();
  int getWidth();
  int getHeight();
  int getFpsNum();
  int getFpsDen();
  int getAspectX();
  int getAspectY();
  int getVideoFrameDuration();

  int getAudioSampleRate();
  int getAudioChannels();
  int getDecodedAudioBytes();
  int getTotalDecodedAudioBytes();
  int getDecodedAudioSampleNum();
  
  int setVideoResampleContext(int , int);

  int isDoneVideo();
  int isDoneAudio();
  
  bool isInterlaced();
  
  void seekToTop();
  
};

#endif
