/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "MP4Handler.h"

/**********************************************************************
CONST & DEST
**********************************************************************/
////////////////////////////////////////
MP4Handler::MP4Handler() {
  _hFile = NULL;
  _filename = NULL;

  _vFrameCount = 0;
  _aFrameCount = 0;
  
  _vDelay = 0;
}

////////////////////////////////////////
MP4Handler::~MP4Handler() {

  if (_filename != NULL) {
    free(_filename);
    _filename = NULL;
  }

  fprintf(stderr, "mp4: object deleted\n");
}

/**********************************************************************
PUBLIC METHOD
**********************************************************************/
////////////////////////////////////////
bool MP4Handler::init() { //FIXME
  _width  = 720;
  _height = 480;
  return false;
}

////////////////////////////////////////
bool MP4Handler::open(const char *mp4file) {
  if(mp4file != NULL) {
    _filename = (char *)malloc(sizeof(char) * MAX_MP4_FILENAME);
    strncpy(_filename, mp4file, 128);
  }

  fprintf(stderr, "mp4: create file %s\n", mp4file);
  _hFile = MP4Create(mp4file, MP4_DETAILS_ERROR, 0);

  return true;
}

////////////////////
bool MP4Handler::open() {

  if (_filename == NULL) {
    _filename = (char *)malloc(sizeof(char) * MAX_MP4_FILENAME);
    sprintf(_filename, "output.mp4");
  }

  fprintf(stderr, "mp4: create file %s\n", _filename);
  _hFile = MP4Create(_filename, MP4_DETAILS_ERROR, 0);

  if (_hFile != NULL) {
    return true;
  } else {
    return false;
  }
}

////////////////////////////////////////
bool MP4Handler::close() {

  optimizeMP4File();

  /*
  fprintf(stderr, "mp4: %s\n",MP4Info(_hFile, _idVideoTrack));
  fprintf(stderr, "mp4: %s\n",MP4Info(_hFile, _idAudioTrack));
  */

  if (_hFile != NULL) {
    MP4Close(_hFile);
    _hFile = NULL;
  }

  fprintf(stderr, "mp4: closed\n");

  return true;
}

////////////////////////////////////////
bool MP4Handler::setMP4TimeScale(int scale) {
  _timeScale = scale;
  MP4SetTimeScale(_hFile, _timeScale);
  return true;
}

////////////////////////////////////////
bool MP4Handler::setVideoDelayTime(int delay) {
	
	if (delay <= 0) {
		_vDelay = 0;
	} else {
		_vDelay = _timeScale / 10 * delay;
	}
	
	return true;
}

////////////////////////////////////////
bool MP4Handler::optimizeMP4File() {
  //MP4Optimize(_filename, "a.mp4", MP4_DETAILS_ERROR );
  return true;
}

////////////////////////////////////////
///// Video /////
////////////////////////////////////////
bool MP4Handler::addH264VideoTrack(u_int8_t *sps, int32_t width, int32_t height,
				      int32_t scale) {
	_width  = width;
	_height = height;
	
	//_width = 854;
	//_height = 480;
  
  MP4SetVideoProfileLevel(_hFile, 0x15);

  _idVideoTrack = MP4AddH264VideoTrack(_hFile,
				       scale,
				       MP4_INVALID_DURATION,
				       _width,
				       _height,
				       sps[1],
				       sps[2],
				       sps[3],
				       3); //FIXME ???

  return true;
}

////////////////////////////////////////
bool MP4Handler::setH264SequenceParameter(u_int8_t *sps, int len) {

  return (MP4AddH264SequenceParameterSet(_hFile,
					 _idVideoTrack,
					 sps, 
					 len));
}

////////////////////////////////////////
bool MP4Handler::setH264PictureParameter(u_int8_t *pps, int len) {

  return (MP4AddH264PictureParameterSet(_hFile,
					_idVideoTrack,
					pps, 
					len));
}

////////////////////////////////////////
bool MP4Handler::addAACAudioTrack(u_int32_t timeScale) {
  _idAudioTrack = MP4AddAudioTrack(_hFile,
				   timeScale, //i.e sample rate
				   MP4_INVALID_DURATION,
				   MP4_MPEG4_AUDIO_TYPE);

  MP4SetAudioProfileLevel(_hFile, 0x0F);

  MP4SetTrackESConfiguration(_hFile,
			     _idAudioTrack,
			     (u_int8_t*)_asc,
			     _ascLength);

  return true;
}

////////////////////////////////////////
bool MP4Handler::addAACAudioTrack(u_int32_t timescale, uint8_t* asc, long unsigned int len) {
  _idAudioTrack = MP4AddAudioTrack(_hFile,
				   timescale, //i.e sample rate
				   MP4_INVALID_DURATION,
				   MP4_MPEG4_AUDIO_TYPE);

  MP4SetAudioProfileLevel(_hFile, 0x0F);

  MP4SetTrackESConfiguration(_hFile,
			     _idAudioTrack,
			     asc,
			     len);

  return true;
}

////////////////////////////////////////
bool MP4Handler::writeMP4VideoSample(u_int8_t *buf, int bytesWritten, u_int64_t duration) {

  u_int64_t d = duration + _vDelay;
  _vDelay = 0;
  
  return (MP4WriteSample(_hFile,
			 _idVideoTrack,
			 buf,
			 bytesWritten,
			 //duration,
			 d,
			 0, // rendering offset
			 0)); //bool isSyncSample DEFAULT(true)));

}

////////////////////////////////////////
// over load
bool MP4Handler::writeMP4VideoSample(u_int8_t *buf, int bytesWritten, u_int64_t duration, bool flag) {
	
  u_int64_t d = duration + _vDelay;
  _vDelay = 0;

  return (MP4WriteSample(_hFile,
						 _idVideoTrack,
						 buf,
						 bytesWritten,
						 //duration,
						 d,
						 0, // rendering offset
						 flag)); //bool isSyncSample DEFAULT(true)));
}

bool MP4Handler::writeMP4VideoSample(u_int8_t *buf, int bytesWritten, u_int64_t duration, int offset, bool flag) {

  u_int64_t d = duration + _vDelay;
  _vDelay = 0;

  //fprintf(stderr, "mp4: vcount %d\n", ++_vFrameCount);
  ++_vFrameCount;
  
  return (MP4WriteSample(_hFile,
			 _idVideoTrack,
			 buf,
			 bytesWritten,
						 //duration,
						 d,
			 offset, // rendering offset
			 flag)); //bool isSyncSample DEFAULT(true)));
}

////////////////////////////////////////
bool MP4Handler::writeMP4AudioSample(u_int8_t *buf, int bytesWritten, u_int64_t duration) {

	//fprintf(stderr, "mp4: duration %lld\n", duration);
  MP4WriteSample(_hFile,
		 _idAudioTrack,
		 buf,
		 bytesWritten,
		 duration,
		 0,
		 false);
  return true;
}

////////////////////////////////////////
unsigned char **MP4Handler::getASCRef() {
  return &_asc;
}

////////////////////////////////////////
long unsigned int *MP4Handler::getASCLengthRef() {
  return &_ascLength;
}

/**********************************************************************
FUNCTION FOR TESTING
**********************************************************************/
#ifdef TEST_MP4
int main() {
  char s[32] = "mp4file.mp4";

  MP4Handler *ctr = new MP4Handler();
  ctr->open(NULL);
  ctr->close();
  delete ctr;
  
}
#endif
