/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <inttypes.h>

#ifdef _USE_MAKE_
#include <x264.h>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#endif // _USE_MAKE_

#include "X264Encoder.h"
#include "FFMPEGDecoder.h"
#include "MP4Handler.h"
#include "FAACEncoder.h"

void encode(void);
void flush(void);
void initCodecs(char* filename);
void closeCodecs(void);

void storeB(u_int8_t*, int, int);
u_int8_t bBuf[160000];
int bBufInd;
int bFrames;
int bSize[5];
int bDuration[5];

void storeP(u_int8_t*, int, int);
u_int8_t pBuf[120000];
int      pSize;
int      pDuration;

#ifndef ONLY_AUDIO
static void copy_frame(AVPicture *, x264_picture_t  *);
#endif

static  bool ctr = true;
static void sigIntHandler(int s) {
  ctr = false;
}

// globals 
X264Encoder   *x264;
x264_picture_t *pic;
FAACEncoder   *faac;
FFMPEGDecoder *dec;
MP4Handler *mp4;

int _systemTimeScale;
int _videoTimeScale;

////////////////////////////////////////
int main(int argc, char *argv[]) {

  pSize = -1;
  bFrames = 0;
  bBufInd = 0;
  
  _systemTimeScale = 90000;
  _videoTimeScale = _systemTimeScale;
 
  if (argc < 2) {
    fprintf(stderr, "usage: Esperance inputfile.m2p\n");
    return EXIT_SUCCESS;
  }
 
  signal(SIGINT, sigIntHandler);

  initCodecs(argv[1]);

  fprintf(stderr, "sys: start encoding\n");
  
  encode();
  
  closeCodecs();

  return EXIT_SUCCESS;
}


////////////////////////////////////////
void encode() {
  int i = 0;
#ifndef ONLY_AUDIO  
  int j = 0;
#endif
#ifndef ONLY_VIDEO  
  int k = 0;
#endif
  int slice = 0;

  for(i = 0;ctr == true;) {

    /////
#ifndef ONLY_VIDEO
    if (faac->isNeedBuffer() == false) {
      faac->encode();
      if (faac->getBytesWritten() > 0) {
	mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
				 faac->getEncodedBytes(),
				 faac->getEncodedSamples() / faac->getNumChannels());
	puts("");
      }
      continue;
    }
#endif    
    /////

#ifdef ONLY_VIDEO
    dec->decodeOnlyVideo();
#elif ONLY_AUDIO    
    dec->decodeOnlyAudio();
#else
    dec->decode();
#endif    

    ///// if EOF stop encoding
    if (dec->isFinished() == true) {
      fprintf(stderr, "dec: EOF done!!\n");
      flush();
      break;
    }
    /////

    /////
    if (dec->isDoneVideo() != 0) {
#ifndef ONLY_AUDIO
      fprintf(stderr, "sys: video processing %d\n", ++j);
      if (dec->getWidth() < 650) {
	copy_frame(dec->getVideoYUVRef(), pic);
      } else {
	copy_frame(dec->getDeinterlacedVideoYUVRef(), pic);
	//copy_frame(dec->getVideoYUVRef(), pic);
      }
      
      pic->i_pts = dec->getPts();
      //x264->encode(dec->getVideoDuration());
      x264->encode();

      if ( x264->isChangedHeader() == true) {

	//mp4->setH264SequenceParameter(x264->getSPSRef(), x264->getSPSLen());
	//mp4->setH264PictureParameter (x264->getPPSRef(), x264->getPPSLen());

      } else if ( x264->getBytesWritten() > 0) {

	
#define X264_TYPE_I             0x0002
#define X264_TYPE_P             0x0003
#define X264_TYPE_BREF          0x0004
#define X264_TYPE_B             0x0005
	slice = x264->getSliceType();
	
	int offset;
	/*
	if (slice == 3) {
	  offset = dec->getVideoFrameDuration() * 3;
	} else {
	  offset = 0;
	}

	mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
				 x264->getTotalBytesWritten(),
				 dec->getVideoFrameDuration(),
				 offset,
				 0); // 90000 / 3003 = 29.97 fps
	*/

	if (slice == 1) { // I

	  if (pSize > 0) {
	    if (bFrames == 0) {
	      offset = 0;
	    } else {
	      offset = dec->getVideoFrameDuration() * (bFrames + 1);
	    }
	    
	    mp4->writeMP4VideoSample(pBuf,
				     pSize,
				     pDuration,
				     offset,
				     0);
	    pSize = 0;
	  }

	  if (bFrames > 0) {
	    int p = 0;
	    
	    for (int ii = 0; ii < bFrames; ii++) {
	      
	      mp4->writeMP4VideoSample(bBuf + p,
				       bSize[ii],
				       bDuration[ii]);
	      p += bSize[ii];
	    }
	    bFrames = 0;
	    bBufInd = 0;
	  }
	  
	  mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
				   x264->getTotalBytesWritten(),
				   dec->getVideoFrameDuration(),
				   0,
				   1);
	  
	} else if (slice == 3) { // P
	  fprintf(stderr, "slice %d\n", slice);
	  
	  if (pSize > 0) {

	    if (bFrames == 0) {
	      offset = 0;
	    } else {
	      offset = dec->getVideoFrameDuration() * (bFrames + 1);
	    }

	    mp4->writeMP4VideoSample(pBuf,
				     pSize,
				     pDuration,
				     offset,
				     0);
	    pSize = 0;
	  }

	  if (bFrames > 0) {
	    int p = 0;
	    
	    for (int ii = 0; ii < bFrames; ii++) {

	      mp4->writeMP4VideoSample(bBuf + p,
				       bSize[ii],
				       bDuration[ii]);
	      p += bSize[ii];
	    }
	    bFrames = 0;
	    bBufInd = 0;
	  }
	  
	  storeP(x264->getNalEncodedBufferRef(),
		 x264->getTotalBytesWritten(),
		 dec->getVideoFrameDuration());
	  
	} else if (slice == 5) { // B
	  storeB(x264->getNalEncodedBufferRef(),
		 x264->getTotalBytesWritten(),
		 dec->getVideoFrameDuration());
	}
      }
      
      puts("");
#endif      
    } else if (dec->isDoneAudio() != 0) { // AUDIO
#ifndef ONLY_VIDEO      
      fprintf(stderr, "sys: audio processing %d\n", k++);
      faac->encode((int32_t *)dec->getAudioRAWRef(),
		   (int32_t  )dec->getDecodedAudioSampleNum() / 2);

      if (faac->getBytesWritten() > 0) {
	mp4->writeMP4AudioSample(faac->getEncodedBufferRef(),
				 faac->getEncodedBytes(),
				 faac->getEncodedSamples() / faac->getNumChannels());
	puts("");
      }
#endif      
    }
    /////
  }
}

////////////////////////////////////////
void flush() {
#ifndef ONLY_AUDIO      
  x264->encode();
  if ( x264->getTotalBytesWritten() > 0) {
    mp4->writeMP4VideoSample(x264->getNalEncodedBufferRef(),
			     x264->getTotalBytesWritten(),
			     dec->getVideoFrameDuration());
    fprintf(stderr, "sys: flush video to avoid last white frame %d\n"
	    , x264->getBytesWritten());
  }
#endif

  return;
}

////////////////////////////////////////
void initCodecs(char* filename) {

  dec  = new FFMPEGDecoder();

  dec->open(filename);
  dec->init();
  //dec->init(1); // testing for seeking function
  
  faac = new FAACEncoder(dec->getAudioSampleRate(),
			 dec->getAudioChannels(),
			 128);
  faac->init();
  faac->open();

  x264 = new X264Encoder();

  x264->open(dec->getWidth(),  dec->getHeight(),
	     dec->getFpsDen(), dec->getFpsNum());

  x264->init();
  pic = x264->getX264PictureRef();
  
  mp4 = new MP4Handler();
  mp4->init();
  mp4->open();
  mp4->setMP4TimeScale(_systemTimeScale);

#ifndef ONLY_AUDIO  
  if (dec->hasVideo() == true) {
    mp4->addH264VideoTrack(x264->getSPSRef(), x264->getWidth(), x264->getHeight(), _videoTimeScale);
    mp4->setH264SequenceParameter(x264->getSPSRef(), x264->getSPSLen());
    mp4->setH264PictureParameter (x264->getPPSRef(), x264->getPPSLen());
  }
#endif  

#ifndef ONLY_VIDEO  
  if (dec->hasAudio() == true) {
    //faac->getDecoderSpecificInfo(mp4->getASCRef(), mp4->getASCLengthRef());
    //mp4->addAACAudioTrack(faac->getSampleRate());
    
    faac->setDecoderSpecificInfo();
    mp4->addAACAudioTrack(faac->getSampleRate(),
			  faac->getASCRef(),
			  faac->getASCLength());
  }
#endif  
}

void closeCodecs() {

  mp4->close();
  x264->close();
  faac->close();
  dec->close();

  delete x264;
  delete faac;
  delete dec;
  delete mp4;
}

#ifndef ONLY_AUDIO
static void copy_frame(AVPicture *pin, x264_picture_t  *pout) {

  /*
  // Works very fine. In case 720 x 480 but,,,
  pout->img.i_stride[0] = pin->linesize[0];
  pout->img.i_stride[1] = pin->linesize[1];
  pout->img.i_stride[2] = pin->linesize[2];
  pout->img.i_stride[3] = pin->linesize[3];

  // Y
  memcpy(pout->img.plane[0], pin->data[0],
	 dec->getHeight() * pin->linesize[0]    );
  
  // U
  memcpy(pout->img.plane[1], pin->data[1],
	 dec->getHeight() * pin->linesize[1] / 2);
  //fprintf(stderr, "linesize U: %d", pin->linesize[1]);

  // V
  memcpy(pout->img.plane[2], pin->data[2],
	 dec->getHeight() * pin->linesize[2] / 2);
  */

  /* ISSUE: double free
  pout->img.plane[0] = pin->data[0];
  pout->img.plane[1] = pin->data[1];
  pout->img.plane[2] = pin->data[2];
  */

  // slow but stable
  int i, x, y;
  int stride0 = pin->linesize[0];
  int stride1 = pin->linesize[1];
  int stride2 = pin->linesize[2];

  for (y = 0, i = 0; y < dec->getHeight(); y++) {
    for (x = 0; x < dec->getWidth(); x++) {
      pout->img.plane[0][i++] = pin->data[0][stride0 * y + x]; // Y
    }
  }

  for (y = 0, i = 0; y < dec->getHeight() / 2; y++) {
    for (x = 0; x < dec->getWidth()/2; x++) {
      pout->img.plane[1][i]   = pin->data[1][stride1 * y + x]; // U
      pout->img.plane[2][i++] = pin->data[2][stride2 * y + x]; // V
    }
  }
  
}
#endif

void storeP(u_int8_t* data, int size, int duration) {
  pDuration = duration;
  pSize = size;
  memcpy(pBuf, data, size);
}

void storeB(u_int8_t* data, int size, int duration) {
  
  bSize[bFrames] = size;
  bDuration[bFrames] = duration;
  memcpy(bBuf + bBufInd, data, size);
  
  bBufInd += size;
  bFrames++;
}
