/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 ********************************************************************************/
 
#include <unistd.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>

//#include <x264.h>

#include "X264Encoder.h"

/**********************************************************************
CONSTRACTOR & DESTRACTOR
**********************************************************************/

////////////////////////////////////////
X264Encoder::X264Encoder() {

  _hEncoder = NULL;
  _param = NULL;

  _sps = NULL;
  _pps = NULL;
  _changeSPS = false;
  _changePPS = false;
  _headerChanged = false;

  _bytesWritten = 0;
  _totalBytesWritten = 0;

  _nalEncodeBufferSize = 1000000;
  _nalEncodeBuffer = (u_int8_t *)malloc(_nalEncodeBufferSize);

}


////////////////////////////////////////
X264Encoder::~X264Encoder() {
  if (_hEncoder != NULL) {
    x264_encoder_close(_hEncoder);
  }
  
  if (_param != NULL) {
    free(_param);
    _param = NULL;
  }

  if (_nalEncodeBuffer != NULL) {
    free(_nalEncodeBuffer);
    _nalEncodeBuffer = NULL;
  }

  if (_sps != NULL) {
    free(_sps);
    _sps = NULL;
  }

  if (_pps != NULL) {
    free(_pps);
    _pps = NULL;
  }
  
  x264_picture_clean(&_picIn);

  fprintf(stderr, "x264: object deleted\n");
}


/**********************************************************************
PRIVATE METHOD
**********************************************************************/

////////////////////////////////////////
void X264Encoder::setSPS(int idx) {

  if (_sps != NULL) {
    free(_sps);
    _sps = NULL;
  }

  _spsLen = 1 + _nal[idx].i_payload;
  _sps = (u_int8_t *)malloc(_spsLen);
  _sps[0] = 0x67;

  memcpy(&_sps[1], _nal[idx].p_payload, _spsLen-1);
}


////////////////////////////////////////
void X264Encoder::setPPS(int idx) {

  if (_pps != NULL) {
    free(_pps);
    _pps = NULL;
  }

  _ppsLen = 1 + _nal[idx].i_payload;
  _pps = (u_int8_t *)malloc(_ppsLen);
  _pps[0] = 0x68;

  memcpy(&_pps[1], _nal[idx].p_payload, _ppsLen-1);
}


////////////////////////////////////////
void X264Encoder::setDefaultParam() { // FIXME
#ifdef __POWERPC__
  _param->cpu = X264_CPU_ALTIVEC;
#endif

  // I-FRAME
  // insert at least one I-FRAME per 1 sec
  //_param->i_keyint_max = 10 * _param->i_fps_den / _param->i_fps_num;

  // QUALITY
  /* 6 is best but slowest*/
  _param->analyse.i_subpel_refine = 2;
  _param->b_repeat_headers = 1;

  /* num of frame reference and motion estimation algo */
  _param->i_frame_reference = 1;
  _param->analyse.i_me_method = 1;

  /* user analyse: i4x4 i8x8 p8x8 p4x4 b8x8 */
  ///_param->analyse.inter = ~0;

  //_param->b_bframe_pyramid = 1;
  
  //_param->analyse.b_bframe_rdo = 1; 
  _param->i_bframe = 2;
  _param->b_bframe_pyramid = 0;
  _param->b_bframe_adaptive = 0;
	
  // QP PARAM (If setted, will be VBR)
  // qp_min = 30 about  600 kbps
  // qp_min = 27 about 1000 kbps
  _param->rc.i_qp_constant = 26;
  
  //_param->rc.i_qp_min = 24;
  //_param->rc.i_qp_max = 36;
  
  //_param->rc.i_rf_constant = 25;
  //_param->rc.i_vbv_max_bitrate = 1000;
  //_param->rc.i_vbv_buffer_size = 1299700;

  //char x264_state_log[] = "x264_state";

  // 2-pass
  //_param->rc.b_stat_write = 1;
  //_param->rc.psz_stat_out = x264_state_log;

  //_param->rc.b_stat_read = 1;
  //_param->rc.psz_stat_in = x264_state_log;
  
  _param->analyse.b_chroma_me = 1;
  //_param->analyse.b_mixed_references = 1;

  // BIT RATE for CBR
  //_param->rc.b_cbr = 1;
  //_param->rc.i_bitrate = 1;
  //_param->rc.f_rate_tolerance = 1.0;

  // LOG
  _param->i_log_level = X264_LOG_DEBUG;  
  //_param->i_log_level = X264_LOG_NONE;

}

/**********************************************************************
PUBLIC METHOD
**********************************************************************/

////////////////////////////////////////
bool X264Encoder::init(esp_x264_params_t *p) {

  _param = (x264_param_t *)malloc(sizeof(x264_param_t));

  if (_param == NULL) {
    return false;
  }

  x264_param_default(_param);

  _width  = p->width;
  _height = p->height;
  _param->i_width  = _width;
  _param->i_height = _height;
  
  _param->vui.i_sar_width  = p->sar_width;
  _param->vui.i_sar_height = p->sar_height;

  // FPS FRAME
  _param->i_fps_den = p->fps_den;
  _param->i_fps_num = p->fps_num;
  
  // Allocate memory for input buffer
  x264_picture_alloc(&_picIn, X264_CSP_I420, _param->i_width, _param->i_height);

  // set parameter for picture
  _picIn.i_type = X264_TYPE_AUTO;
  _picIn.i_qpplus1 = 0; //???: what is this?

 //CALL: private method //FIXME: order
#ifdef __POWERPC__
  _param->cpu = X264_CPU_ALTIVEC;
#endif

  _param->i_threads = p->threads;

  // I-FRAME
  // insert at least one I-FRAME per 1 sec
  _param->i_keyint_max = p->key_max * _param->i_fps_den / _param->i_fps_num;
  
  // noise reduction
  _param->analyse.i_noise_reduction = p->nr;
  
  _param->analyse.i_direct_mv_pred = X264_DIRECT_PRED_AUTO;
  //_param->analyse.i_direct_mv_pred = 1;

  // QUALITY
  /* 7 is best but slowest*/
  _param->analyse.i_subpel_refine = p->subpel_refine;
  
  if (X264_BUILD < 45 && p->subpel_refine == 7) {
	  _param->analyse.i_subpel_refine = 6;
  }
  
  _param->b_bframe_adaptive = p->bframe_adaptive;
  _param->analyse.b_psnr      = p->psnr;
  _param->analyse.b_weighted_bipred = p->weighted_bipred;
  _param->analyse.b_bidir_me = p->bidir_me;
  _param->analyse.b_fast_pskip = p->fast_pskip;
    
  //_param->analyse.b_transform_8x8   = p->transform_8x8;
  
  /* num of frame reference and motion estimation algo */
  _param->i_frame_reference = p->frame_reference;
  _param->analyse.i_me_method = p->me_method;

  /* user analyse: i4x4 i8x8 p8x8 p4x4 b8x8 */
  //_param->analyse.inter = ~0;

  // QP PARAM (If setted, will be VBR)
  
  switch(p->e_mode) {

	  case X264_QMODE_CRF:
	  
		  _param->rc.i_rf_constant = p->crf;
		  //_param->rc.i_vbv_max_bitrate = 800;
		  //_param->rc.i_vbv_buffer_size = 10;
		  //_param->rc.f_vbv_buffer_init = 1.0;
		  
		  _param->rc.i_qp_min = p->qp_min;
		  _param->rc.i_qp_max = p->qp_max;
		  _param->rc.i_qp_step = p->qp_step;
		  //_param->rc.f_rate_tolerance = 1.0;

		  break;
	  case X264_QMODE_CBR:
		  _param->rc.b_cbr = 1;
		  //_param->rc.i_bitrate = p->cbr / 1000;
		  _param->rc.i_bitrate = 1;
		  break;
	  case X264_QMODE_CQP:
		  _param->rc.i_qp_constant = p->cqp;
		  _param->rc.i_qp_min = p->cqp;
		  _param->rc.i_qp_max = p->cqp;
		  break;
  }
  
  /*
  char x264_state_log[512];// = "/tmp/x264_1st_pass_log";
  char x264_state_log2[512];// = "/tmp/x264_1st_pass_log";
  const char fname[] = "/x264_1st_pass_log";

  //getcwd(x264_state_log2,sizeof(x264_state_log2));
  
  strcat(x264_state_log2,getenv("HOME"));
  
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  
  strcat(x264_state_log2, fname);
  
  fprintf(stderr, "file name : %s\n", x264_state_log2);
  
  sprintf(x264_state_log, "%s", x264_state_log2);
  */
  
  // 2-pass 
  
  // be care  
  // 2 means 1st pass of 2
  // 1 means 2nd pass of 2
  // 0 reserved for non-multipass
  if (p->pass == 2) {
	  _param->rc.b_stat_write = 1;
	  _param->rc.psz_stat_out = p->log_file;
  } else if (p->pass == 1) {
	  _param->rc.b_stat_read = 1;
	  _param->rc.psz_stat_in = p->log_file;
  }
  
  //_param->analyse.b_bframe_rdo = 1; 
  _param->i_bframe = p->bframe;
  //_param->b_bframe_pyramid = 0;
  
  if (p->df_alpha != 0 || p->df_beta != 0) {
	  _param->b_deblocking_filter = 1;
	  _param->i_deblocking_filter_alphac0 = p->df_alpha;
	  _param->i_deblocking_filter_beta = p->df_beta;
  } else {
	  _param->b_deblocking_filter = 1;
  }

  // LOG
  _param->i_log_level = X264_LOG_DEBUG;  
  //_param->i_log_level = X264_LOG_NONE;
	
  return true;
}


////////////////////////////////////////
bool X264Encoder::init() {
  return true;
}


////////////////////////////////////////
bool X264Encoder::open() {

  _hEncoder = x264_encoder_open(_param);
  if (_hEncoder == NULL) {
    fprintf(stderr, "x264: opening codec failure\n");
    return false;
  }

  x264_encoder_reconfig(_hEncoder, _param);  
  x264_encoder_headers(_hEncoder, &_nal, &_nalCount);

  setSPS(1);
  setPPS(2);

  fprintf(stderr, "x264: open codec and reconfig\n");
  
  return true;
}


////////////////////////////////////////
bool X264Encoder::open(int32_t width, int32_t height,
		       int32_t f_den, int32_t f_num) {

  _param = (x264_param_t *)malloc(sizeof(x264_param_t));

  if (_param == NULL) {
    return false;
  }
 
  x264_param_default(_param);

  // SIZE
  _width  = width;
  _height = height;
  _param->i_width  = _width;
  _param->i_height = _height;

  // FPS FRAME
  _param->i_fps_den = f_den;
  _param->i_fps_num = f_num;

  int stride = 0;
  /*
  switch (_param->i_width) {
  case 352:
    stride = 32;
  case 640:
    stride = 0;
  case 704:
    stride = 0;
  case 720:
    stride = 0;
  default:
    stride = 0;
  }
  */

  // Allocate memory for input buffer
  x264_picture_alloc(&_picIn, X264_CSP_I420, _param->i_width + stride, _param->i_height);

  // set parameter for picture
  _picIn.i_type = X264_TYPE_AUTO;
  _picIn.i_qpplus1 = 0; //???: what is this?

 //CALL: private method //FIXME: order
  setDefaultParam();

  _hEncoder = x264_encoder_open(_param);
  if (_hEncoder == NULL) {
    fprintf(stderr, "x264: opening codec failure\n");
    return false;
  }

  x264_encoder_reconfig(_hEncoder, _param);  
  x264_encoder_headers(_hEncoder, &_nal, &_nalCount);

  setSPS(1);
  setPPS(2);

  fprintf(stderr, "x264: open codec and reconfig\n");

  return true;
}


////////////////////////////////////////
bool X264Encoder::close() {
  if (_hEncoder != NULL) {
    x264_encoder_close(_hEncoder);
    _hEncoder = NULL;
  }

  fprintf(stderr, "x264: codec closed\n");

  return true;
}


////////////////////////////////////////
bool X264Encoder::encode(int num) {

  _param->i_fps_num = num;
  x264_encoder_reconfig(_hEncoder, _param);

  return (encode());
}

////////////////////////////////////////
bool X264Encoder::flush() {

  // for nal encode. don't touch in following sequence.
  setTotalBytesWrittenZero();

  _bytesWritten = x264_encoder_encode(_hEncoder,
				      &_nal,
				      &_nalCount,
				      NULL,
				      &_picOut);
  
  if (_bytesWritten < 0) {
    return false;
  }

  return (encodeNal());
}
  
////////////////////////////////////////
bool X264Encoder::encode() {

  // for nal encode. don't touch in following sequence.
  setTotalBytesWrittenZero();

  _bytesWritten = x264_encoder_encode(_hEncoder,
				      &_nal,
				      &_nalCount,
				      &_picIn,
				      &_picOut);
  
  if (_bytesWritten < 0) {
    return false;
  }

  return (encodeNal());
}


// TEST METHOD: near future this will be main method
////////////////////////////////////////
bool X264Encoder::encodeNal() {
  const int b_annexeb = 1; // using long synch work
  int type;
  _bytesWritten = 0;
  
  for (int i = 0; i < _nalCount; i++) {

    _bytesWritten = x264_nal_encode(_nalEncodeBuffer+_totalBytesWritten,
				    &_nalEncodeBufferSize,
				    b_annexeb, // must be 1
				    &_nal[i]);

    type = _nalEncodeBuffer[_totalBytesWritten + 4] & 0x1F;

    //TODO everything
    switch(type) {

    case 0x07: //SPS
      fprintf(stderr, "x264: SPS Header %d\n", _nalCount);
      //setSPS(i);
      //_changeSPS = true;
      //_headerChanged = true;
      break;

    case 0x08: //PPS
      fprintf(stderr, "x264: PPS Header\n");
      //setPPS(i);
      //_changePPS = true;
      //_headerChanged = true;
      break;

    case 0x01:
    case 0x05:
    case 0x06:
      //DONT TEST
      _nalEncodeBuffer[_totalBytesWritten + 0] = ((_bytesWritten - 4) >> 24) & 0xFF;
      _nalEncodeBuffer[_totalBytesWritten + 1] = ((_bytesWritten - 4) >> 16) & 0xFF;
      _nalEncodeBuffer[_totalBytesWritten + 2] = ((_bytesWritten - 4) >>  8) & 0xFF;
      _nalEncodeBuffer[_totalBytesWritten + 3] = ((_bytesWritten - 4) >>  0) & 0xFF;
      _totalBytesWritten += _bytesWritten;
      break;
    }
  }
  
  //fprintf(stderr, "pts %lld \n", _picOut.i_pts);

  return true;
}  
/************************************************************
 * GET
 ************************************************************/

////////////////////////////////////////
x264_picture_t *X264Encoder::getX264PictureRef() {
  return &_picIn;
}


////////////////////////////////////////
u_int8_t *X264Encoder::getNalEncodedBufferRef() {
  return _nalEncodeBuffer;
}


////////////////////////////////////////
int X264Encoder::getBytesWritten() {
  return _bytesWritten;
}


////////////////////////////////////////
int64_t X264Encoder::getTotalBytesWritten() {
  return _totalBytesWritten;
}


////////////////////////////////////////
int X264Encoder::getNalCount() {
  return _nalCount;
}


////////////////////////////////////////
int X264Encoder::getWidth() {
  return _width;
}


////////////////////////////////////////
int X264Encoder::getHeight() {
  return _height;
}


////////////////////////////////////////
u_int8_t *X264Encoder::getSPSRef() {
  _changeSPS = false;
  return _sps;
}


////////////////////////////////////////
int X264Encoder::getSPSLen() {
  return _spsLen;
}


////////////////////////////////////////
u_int8_t *X264Encoder::getPPSRef() {
  _changePPS = false;
  return _pps;
}


////////////////////////////////////////
int X264Encoder::getPPSLen() {
  return _ppsLen;
}


////////////////////////////////////////
bool X264Encoder::isChangedHeader() {
  _headerChanged = false;
  return _headerChanged;
}

u_int8_t X264Encoder::getSliceType() {
  return _picOut.i_type;
}

u_int64_t X264Encoder::getPTS() {
  return _picOut.i_pts;
}


/************************************************************
 * SET
 ************************************************************/

////////////////////////////////////////
void X264Encoder::setWidth(int width) {
  _width = width;
  _param->i_width = _width;
}


////////////////////////////////////////
void X264Encoder::setHeight(int height) {
  _height = height;
  _param->i_height = _height;
}


////////////////////////////////////////
void X264Encoder::setFrameRateFactor(int num, int den) {
  _num = num;
  _den = den;

  _param->i_fps_num = _num;
  _param->i_fps_num = _den;
}


////////////////////////////////////////
void X264Encoder::setTotalBytesWrittenZero() {
  _totalBytesWritten = 0;
}

/**********************************************************************
TEST FUNCTION
**********************************************************************/
/*
int main() {
  X264Encoder *x = new X264Encoder();
  x->init();
  x->open();
  x->close();
  delete x;

  return EXIT_SUCCESS;
}
*/
