import java.io.*;
import sun.misc.Signal;
import sun.misc.SignalHandler;

public class EsperanceJNI {

    private static final String DEFAULT_INPUT_FILENAME = "/Users/Pipin/Desktop/jib2.mpg";
    //private static final String DEFAULT_INPUT_FILENAME = "/Users/Pipin/Development/Visual/resources/hachi.m2p";
    private RefHandler _hMP4;
    private FFMPEGDecoderJNI _ffmpeg;
    private X264EncoderJNI   _x264;
    private FAACEncoderJNI   _faac;
    private MP4HandlerJNI    _mp4;

    private boolean _working;


    public EsperanceJNI() {

	//_hMP4 = new RefHandler("MP4HandlerJNI");
	
	_ffmpeg = new FFMPEGDecoderJNI();
	_x264   = new X264EncoderJNI();
	_faac   = new FAACEncoderJNI();
	_mp4    = new MP4HandlerJNI();

	_working = true;
    }

    /*
    public void perform(Prototype p) {
	EsperanceJNI esp = new EsperanceJNI();

	ESPBreaker breaker = new ESPBreaker(esp);
	//Runtime.getRuntime().addShutdownHook(breaker);

	ESPSignalHandler _sig = new ESPSignalHandler(esp);
	_sig.install();

	esp.openCodecs();

	try {
	    breaker.start();
	} catch (Exception e) {
	    ;
	}

	esp.encode();
	esp.closeCodecs();

	breaker._cnt = false;

	try {
	    breaker.join();
	} catch (Exception e) {
	    ;
	}

	System.out.println("DONE DONE DONE DONE DONE");
    }
    */
    
    protected void openCodecs() {

	byte[] sps = new byte[128];
	byte[] pps = new byte[128];
	byte[] asc = new byte[256];
	int sps_len = 0;
	int pps_len = 0;

	_ffmpeg.alloc();
	_ffmpeg.open(DEFAULT_INPUT_FILENAME);
	_ffmpeg.init();

	_x264.alloc();
	_x264.open(_ffmpeg.getWidth() + 128, _ffmpeg.getHeight(), _ffmpeg.getFpsDen(), _ffmpeg.getFpsNum()); //TODO
	_x264.init();
	sps_len = _x264.getSPSLen();
	sps = _x264.getSPSRef(sps);
	pps_len = _x264.getPPSLen();
	pps = _x264.getPPSRef(pps);

	_faac.alloc();
	_faac.init(_ffmpeg.getAudioSampleRate(),
		   _ffmpeg.getAudioChannels());
	_faac.open();

	_mp4.alloc();
	_mp4.open();
	_mp4.init();

	_mp4.addH264VideoTrack(sps, _ffmpeg.getWidth(), _ffmpeg.getHeight(), 90000);
	_mp4.setH264SequenceParameter(sps, sps_len);
	_mp4.setH264PictureParameter(pps, pps_len);

	_faac.setDecoderSpecificInfo();
	_mp4.addAACAudioTrack(_ffmpeg.getAudioSampleRate(),
			      _faac.getASC(asc),
			      _faac.getASCLength());
    }

    
    protected void closeCodecs() {

	_mp4.close();
	_mp4.delete();

	_ffmpeg.close();
	_ffmpeg.delete();

	_x264.close();
	_x264.delete();
	
	_faac.close();
	_faac.delete();
    }

    protected void encode() {
	int diff = 0;
	
	byte[] bufX264  = new byte[160000];
	byte[] bufFAAC  = new byte[18192];
	byte[] bufAudio = new byte[40000];

	if (_ffmpeg.getWidth() == 640) {
	    diff = 32;
	} else if (_ffmpeg.getWidth() == 720) {
	    diff = 48;
	}
	
	byte[] y = new byte[(_ffmpeg.getWidth() + diff) * _ffmpeg.getHeight()];
	int ysize = 0;
	byte[] u = new byte[((_ffmpeg.getWidth() + diff) / 2) * _ffmpeg.getHeight()];
	int usize = 0;
	byte[] v = new byte[((_ffmpeg.getWidth() + diff) / 2) * _ffmpeg.getHeight()];
	int vsize = 0;
	
	for (; _ffmpeg.isFinished() != true && this._working == true; ) {

	    // if need more buffer for audio
	    if (_faac.needBuffer() == false) {
		
		_faac.encode();
		
		// write to mp4 file
		if (_faac.getBytesWritten() > 10) {
		    _mp4.writeMP4AudioSample(_faac.getEncodedBufferRef(bufFAAC),
					     _faac.getEncodedBytes(),
					     _faac.getEncodedSamples() / 2);
		}
	    }

	    _ffmpeg.decode();

	    if (_ffmpeg.isFinished() == true) {
		System.out.println("sys: done encoding");
		break;
	    }

	    if (_ffmpeg.doneVideo() != 0) {
		
		// copy frame		
		ysize = _ffmpeg.getVideoY(y);
		usize = _ffmpeg.getVideoU(u);
		vsize = _ffmpeg.getVideoV(v);

		// x264 encode
		_x264.encode(y, u, v, ysize, usize, vsize);
		
		// write to mp4 file
		if (_x264.getBytesWritten() > 0) {
		    //System.err.println("encoded" +  _x264.getBytesWritten());

		    _mp4.writeMP4VideoSample(_x264.getNalEncodedBufferRef(bufX264), _x264.getBytesWritten(), _ffmpeg.getVideoFrameDuration());
		}
	    } else if (_ffmpeg.doneAudio() != 0) {
		// faac.encode
		//System.err.println("Encode Audio");
		_faac.encode(_ffmpeg.getAudioRAWRef(bufAudio),
			     _ffmpeg.getDecodedAudioSampleNum() / 2);
		
		// write to mp4 file
		if (_faac.getBytesWritten() > 10) {
		    _mp4.writeMP4AudioSample(_faac.getEncodedBufferRef(bufFAAC),
					     _faac.getEncodedBytes(),
					     _faac.getEncodedSamples() / 2);
		}

	    }
	}
    }

    public void stop() {
	System.out.println("stop");
	_working = false;
    }

    public void perform() {

    }
    
    public static void main(String[] args) {
	EsperanceJNI esp = new EsperanceJNI();

	ESPBreaker breaker = new ESPBreaker(esp);
	//Runtime.getRuntime().addShutdownHook(breaker);

	ESPSignalHandler _sig = new ESPSignalHandler(esp);
	_sig.install();

	esp.openCodecs();

	/*
	try {
	    breaker.start();
	} catch (Exception e) {
	    ;
	}
	*/

	esp.encode();

	esp.closeCodecs();

	/*
	breaker._cnt = false;

	try {
	    breaker.join();
	} catch (Exception e) {
	    ;
	}
	*/

	System.out.println("DONE DONE DONE DONE DONE");
    }

}

class ESPBreaker extends Thread {

    public boolean _cnt;
    EsperanceJNI _esp;
    BufferedReader _input;

    public ESPBreaker(EsperanceJNI esp) {
	_esp = esp;
	_cnt = true;
    }

    public void run() {

	String line = null;
	
	System.out.println("stop encoding: s");

	_input = new BufferedReader(new InputStreamReader(System.in));

	//_esp.stop();
	for (;_cnt == true;) {

	    try {
		line = _input.readLine();
	    } catch (Exception e) {
		;
	    }

	    if (line.equals("s")) {
		_esp.stop();
		break;
	    }
	}

	System.out.println("ESP Thread terminated");
	
    }
}

class ESPSignalHandler implements SignalHandler {
    private EsperanceJNI _esp;
    private SignalHandler _hSig;

    public ESPSignalHandler(EsperanceJNI esp) {
	_esp = esp;
    }

    public void install() {

	Signal sig = new Signal("INT");

	Signal.handle(sig, this);
    }

    public void handle(Signal sig) {
	_esp.stop();
    }
}