/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../FFMPEGDecoder.h"
#include "FFMPEGDecoderJNI.h"

FFMPEGDecoder *ffmpeg;

/*
 *
 */
JNIEXPORT void JNICALL Java_FFMPEGDecoderJNI_alloc
(JNIEnv *env, jobject obj) {
  
  ffmpeg = new FFMPEGDecoder();
  
}


/*
 *
 */
JNIEXPORT void JNICALL Java_FFMPEGDecoderJNI_delete
(JNIEnv *env, jobject obj) {
  
  delete ffmpeg;
  
}


/*
 *
 */
JNIEXPORT jboolean JNICALL Java_FFMPEGDecoderJNI_open
(JNIEnv *env, jobject, jstring filename) {

  jboolean bl;

  char *inputfile;
  const char *s;

  int file_name_length = env->GetStringLength(filename);

  if (file_name_length >= 256) {
    return false;
  }
  
  inputfile = (char *)malloc(file_name_length + 1);

  s = env->GetStringUTFChars(filename, NULL);

  memcpy(inputfile, s, file_name_length);
  inputfile[file_name_length] = '\0';


  ffmpeg = new FFMPEGDecoder();

  bl = ffmpeg->open(inputfile);

  env->ReleaseStringUTFChars(filename, s);
  free(inputfile);

  return bl;
}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_FFMPEGDecoderJNI_init
(JNIEnv *env, jobject obj) {

  return (jboolean)ffmpeg->init();

}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_FFMPEGDecoderJNI_close
(JNIEnv *env, jobject obj) {

  return (jboolean)ffmpeg->close();

}


/*
 *
 */
JNIEXPORT jboolean JNICALL Java_FFMPEGDecoderJNI_decode
(JNIEnv *env, jobject obj) {

  return (jboolean)ffmpeg->decode();

}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getWidth
(JNIEnv *env, jobject obj) {
  
  return (jint)ffmpeg->getWidth();
  
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getHeight
(JNIEnv *evn, jobject obj) {
  
  return (jint)ffmpeg->getHeight();
  
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getFpsNum
(JNIEnv *env, jobject obj) {
  
  return (jint)ffmpeg->getFpsNum();

}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getFpsDen
(JNIEnv *env, jobject obj) {

  return (jint)ffmpeg->getFpsDen();

}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getVideoFrameDuration
(JNIEnv *env, jobject obj) {

  return (jint)ffmpeg->getVideoFrameDuration();

}

/*
 *
 */
JNIEXPORT void JNICALL Java_FFMPEGDecoderJNI_getVideoYUV
(JNIEnv *env, jobject obj, jbyteArray buf) {
  fprintf(stderr, "hoghoge");
  env->SetByteArrayRegion(buf,     0, 256, (jbyte*)ffmpeg->getVideoYUVRef());
  fprintf(stderr, "hoghoge");
  env->SetByteArrayRegion(buf+256, 0, 256, (jbyte*)ffmpeg->getVideoYUVRef());
  fprintf(stderr, "hoghoge");
  env->SetByteArrayRegion(buf+512, 0, 256, (jbyte*)ffmpeg->getVideoYUVRef());
  fprintf(stderr, "hoghoge");
}

JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getVideoY
(JNIEnv *env, jobject obj, jbyteArray buf) {
  int linesize = ffmpeg->getVideoYSize() * ffmpeg->getHeight();
  
  env->SetByteArrayRegion(buf, 0, linesize, (jbyte*)ffmpeg->getVideoYRef());

  return ffmpeg->getVideoYSize();
}

JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getVideoU
(JNIEnv *env, jobject obj, jbyteArray buf) {
  int linesize = ffmpeg->getVideoUSize() * ffmpeg->getHeight();;

  env->SetByteArrayRegion(buf, 0, linesize, (jbyte*)ffmpeg->getVideoURef());

  return ffmpeg->getVideoUSize();
}

JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getVideoV
(JNIEnv *env, jobject obj, jbyteArray buf) {
  int linesize = ffmpeg->getVideoVSize() * ffmpeg->getHeight();

  env->SetByteArrayRegion(buf, 0, linesize / 2, (jbyte*)ffmpeg->getVideoVRef());

  return ffmpeg->getVideoVSize();
}

/*
 *
 */
JNIEXPORT void JNICALL Java_FFMPEGDecoderJNI_getDeinterlacedVideoYUV
(JNIEnv *env, jobject obj, jbyteArray buf){
  env->SetByteArrayRegion(buf, 0, 4, (jbyte*)ffmpeg->getDeinterlacedVideoYUVRef());
}

/*
 *
 */
JNIEXPORT jbyteArray JNICALL Java_FFMPEGDecoderJNI_getAudioRAWRef
(JNIEnv *env, jobject obj, jbyteArray buf) {

  env->SetByteArrayRegion(buf, 0, ffmpeg->getDecodedAudioSampleNum()*2, (jbyte*)ffmpeg->getAudioRAWRef());
  
  return buf;
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getDecodedAudioSampleNum
(JNIEnv *env, jobject obj) {
  return ffmpeg->getDecodedAudioSampleNum();
}


/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getAudioSampleRate
(JNIEnv *env, jobject obj) {
    return (ffmpeg->getAudioSampleRate());
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_getAudioChannels
(JNIEnv *env, jobject obj) {
  return (ffmpeg->getAudioChannels());
}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_FFMPEGDecoderJNI_isFinished
(JNIEnv *env, jobject obj) {

  return (ffmpeg->isFinished());

}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_doneVideo
(JNIEnv *env, jobject obj) {
  return (ffmpeg->isDoneVideo());
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_FFMPEGDecoderJNI_doneAudio
(JNIEnv *env, jobject obj) {
  return (ffmpeg->isDoneAudio());
}
