/********************************************************************************
Copyright (C) Takuya.S

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../X264Encoder.h"
#include "X264EncoderJNI.h"

X264Encoder *x264;
x264_picture_t* pic;

/*
 *
 */
JNIEXPORT void JNICALL Java_X264EncoderJNI_alloc
(JNIEnv *env, jobject obj) {

  x264 = new X264Encoder();
  
  pic = x264->getX264PictureRef();
}


/*
 *
 */
JNIEXPORT void JNICALL Java_X264EncoderJNI_delete
(JNIEnv *env, jobject obj) {

  delete x264;

}


/*
 *
 */
JNIEXPORT jboolean JNICALL Java_X264EncoderJNI_open
(JNIEnv *env, jobject, jint width, jint height, jint fps_den, jint fps_num) {

  return (x264->open(width, height, fps_den, fps_num));

}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_X264EncoderJNI_init
(JNIEnv *env, jobject obj) {

  return (x264->init());

}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_X264EncoderJNI_close
(JNIEnv *env, jobject obj) {

  return (x264->close());

}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_X264EncoderJNI_encode
(JNIEnv *env, jobject obj) {

  return (x264->encode());

}

/*
 *
 */
JNIEXPORT jboolean JNICALL Java_X264EncoderJNI_encode___3B_3B_3BIII
(JNIEnv *env, jobject obj, jbyteArray y, jbyteArray u, jbyteArray v, jint ys, jint us, jint vs) {
  
  int height = x264->getHeight();
  
  jbyte ny[ys * height];
  jbyte nu[us * height];
  jbyte nv[vs * height];

  pic->img.i_stride[0] = ys;
  pic->img.i_stride[1] = us;
  pic->img.i_stride[2] = vs;
  pic->img.i_stride[3] = vs;

  env->GetByteArrayRegion(y, 0, ys * height , ny);
  env->GetByteArrayRegion(u, 0, us * height , nu);
  env->GetByteArrayRegion(v, 0, vs * height / 2 , nv);
  
  memcpy(pic->img.plane[0], ny, ys * height);
  memcpy(pic->img.plane[1], nu, us * height);
  memcpy(pic->img.plane[2], nv, vs * height / 2);
  
  return (x264->encode());
  
}

/*
 *
 */
JNIEXPORT jbyteArray JNICALL Java_X264EncoderJNI_getSPSRef
(JNIEnv *env, jobject obj, jbyteArray array) {

  unsigned char *buf;
  int bytes;

  buf   = x264->getSPSRef();
  bytes = x264->getSPSLen();

  env->SetByteArrayRegion(array, 0, bytes, (jbyte*)buf);

  return array;

}


/*
 *
 */
JNIEXPORT jint JNICALL Java_X264EncoderJNI_getSPSLen
(JNIEnv *evn, jobject obj) {
  return x264->getSPSLen();
}


/*
 *
 */
JNIEXPORT jbyteArray JNICALL Java_X264EncoderJNI_getPPSRef
(JNIEnv *env, jobject obj, jbyteArray array) {
  
  unsigned char *buf;
  int bytes;
  
  buf   = x264->getPPSRef();
  bytes = x264->getPPSLen();
  
  env->SetByteArrayRegion(array, 0, bytes, (jbyte*)buf);
  
  return array;

}

/*
 *
 */
JNIEXPORT jint JNICALL Java_X264EncoderJNI_getPPSLen
(JNIEnv *env, jobject obj) {
  return x264->getPPSLen();
}

/*
 *
 */
JNIEXPORT jbyteArray JNICALL Java_X264EncoderJNI_getNalEncodedBufferRef
(JNIEnv *env, jobject obj, jbyteArray buf) {

  env->SetByteArrayRegion(buf, 0, x264->getBytesWritten(),
			  (jbyte*)x264->getNalEncodedBufferRef());

  return buf;
}

/*
 *
 */
JNIEXPORT jint JNICALL Java_X264EncoderJNI_getBytesWritten
(JNIEnv *env, jobject obj) {
  
  return x264->getBytesWritten();
}

