import java.lang.reflect.*;
import java.util.*;
import java.nio.*;
import java.io.*;

import net.jxta.impl.shell.*;
import net.jxta.exception.*;
import net.jxta.util.*;

import net.jxta.peergroup.*;
import net.jxta.discovery.*;
import net.jxta.document.*;
import net.jxta.protocol.*;
import net.jxta.socket.*;
import net.jxta.pipe.*;
import net.jxta.id.*;

public class Prototype extends ShellApp {
    
    private int    _peerMode = -1;
    public static final int ESP_MODE_MAIN = 0;
    public static final int ESP_MODE_MP   = 1;
    public static final int ESP_MODE_HP   = 2;
    public static final int ESP_MODE_GD   = 3;
    
    private ShellEnv    _shellEnv = null;
    private ShellObject _shellObj = null;
    private PeerGroup         _peerGroup = null;
    private DiscoveryService  _discoveryService = null;

    private String _myName    = null;
    private String _className = null;

    private PipeService       _pipeService = null;
    private PipeAdvertisement _pipeAdv = null;
    private InputPipe         _inputPipe = null;
    private OutputPipe        _outputPipe = null;

    private JxtaServerSocket _jxtaServerSocket = null;
    private JxtaSocket       _jxtaSocket       = null;
    private InputStream      _input  = null;
    private OutputStream     _output = null;

    private JxtaServerSocketListenerThread _jsListenThread = null;
    private JxtaSocketReaderThread         _jsReaderThread = null;
    private ClassExecuterThread            _clsExecThread  = null;

    private StructuredTextDocument _returnDoc;
    private boolean _responded = false;
    private Vector _dataVector = null;
    
    /************************************************************
     * Entering Point
     ************************************************************/
    public int startApp(String[] args) {

	initEnv();

	this.flushPipeAdvertisement();
	
	//this.getLocalAdvertisementEnum();
	//this.getRemoteAdvertisement(2);

	getOpt(args);

	activate();
	
	delete();
	
	return 0;
    }

    public void stopApp() {

    }

    /************************************************************
     * Main Controll Routine
     ************************************************************/
    private boolean activate() {
	
	this.createPipeAdvertisement(_myName);
	this.publishPipeAdvertisementLocal();

	if (_peerMode == ESP_MODE_MAIN) {
	    // main peer
	    
	    Thread readerT = null;
	    
	    if (_className == null) {
		System.err.println("Please set class name: -cHogeClass ");
		return false;
	    }

	    getRemoteAdvertisement(2); // 2:OTHERE
	    _jxtaSocket = findAvailableJxtaSocket(getLocalAdvertisementEnum());

	    if (_jxtaSocket == null) {
		return false;
	    }

	    this.idleState(5);
	    
	    this._jsReaderThread = new JxtaSocketReaderThread(this, _jxtaSocket);

	    readerT = new Thread(this._jsReaderThread);
	    
	    this._clsExecThread = new ClassExecuterThread(loadClassByName(_className),
							  this);

	    Thread classT = new Thread(this._clsExecThread);
	    
	    try {
		classT.start();
		classT.join();
	    } catch (Exception e) {
	    }

	    this.idleState(40);

	    _clsExecThread.die();
	    _jsReaderThread.die();
	
	    
	} else if (_peerMode == ESP_MODE_MP) {
	    // act as a module provider

	    Thread socketT = null;
	    Thread readerT = null;

	    this.publishPipeAdvertisementRemote();

	    try {
		this._jxtaServerSocket = new JxtaServerSocket();
		this._jxtaServerSocket.setSoTimeout(30 * 1000);
	    } catch (IOException e) {
		System.err.println(e.getMessage());
		return false;
	    }

	    // Listener Thread
	    this._jsListenThread = new JxtaServerSocketListenerThread(this,
								      _jxtaServerSocket,
								      _peerGroup,
								      _pipeAdv);
	    socketT = new Thread(this._jsListenThread);

	    try {
		socketT.start();
	    } catch (Exception e) {
		e.getMessage();
	    }
	    
	    this.idleState(10);
	    
	    //
	    // main sequence
	    //
	    
	    // Reader Thread
	    if (_jsListenThread.isConnected() == true) {
		_jxtaSocket = _jsListenThread.getSocket();
	    } else {
		return false;
	    }

	    this._jsReaderThread = new JxtaSocketReaderThread(this, _jxtaSocket);

	    readerT = new Thread(this._jsReaderThread);

	    try {
		readerT.start();
	    } catch (Exception e) {
		e.getMessage();
	    }
	    

	    //
	    // end
	    //

	    this.idleState(50);

	    _jsListenThread.die();
	    _jsReaderThread.die();
	    
	    try {
		System.err.println("Join Server Socket Thread");
		socketT.join(3 * 60 * 1000);
	    } catch (InterruptedException e) {
		System.out.println(e.getMessage());
	    }
	    
	} else if (_peerMode == ESP_MODE_HP) {
	    Thread socketT = null;
	    Thread readerT = null;
	    
	    this._clsExecThread = new ClassExecuterThread(loadClassByName(_className));
	    this._clsExecThread.init(this);
	    
	    // act as a host provider
	    this.publishPipeAdvertisementRemote();

	    try {
		this._jxtaServerSocket = new JxtaServerSocket();
		this._jxtaServerSocket.setSoTimeout(60 * 1000);
	    } catch (IOException e) {
		System.err.println(e.getMessage());
		return false;
	    }
	    
	    this._jsListenThread = new JxtaServerSocketListenerThread(this,
								      _jxtaServerSocket,
								      _peerGroup,
								      _pipeAdv);
	    socketT = new Thread(this._jsListenThread);

	    try {
		socketT.start();
	    } catch (Exception e) {
		e.getMessage();
	    }

	    //
	    // main sequence
	    //
	    
	    // Reader Thread
	    while (_jsListenThread.isConnected() == false) {
		this.idleState(1);
	    }
	    _jxtaSocket = _jsListenThread.getSocket();

	    this._jsReaderThread = new JxtaSocketReaderThread(this, _jxtaSocket);
	    readerT = new Thread(this._jsReaderThread);

	    try {
		readerT.start();
	    } catch (Exception e) {
		e.getMessage();
	    }

	    //
	    // end
	    //
	    
	    this.idleState(20 * 60);	    
	    _jsListenThread.die();
	    _jsReaderThread.die();
	    
	    try {
		System.err.println("Join Server Socket Thread");
		socketT.join(3 * 60 * 1000);
	    } catch (InterruptedException e) {
		System.out.println(e.getMessage());
	    }

	    
	} else if (_peerMode == ESP_MODE_GD) {
	    
	}

	return true;
    }

    public void handleXMLDocument(byte[] data) {
	StructuredTextDocument std = null;

	try {
	    std = (StructuredTextDocument)DocumentUtilities.getStructuredXmlDocument(data);
	} catch (JxtaException e) {
	    System.err.println(e.getMessage());
	}

	this.handleXMLDocument(std);
    }

    public void handleXMLDocument(StructuredTextDocument doc) {
	
	Enumeration enu = null;
	TextElement ele = null;
	String name = null;
	String  val = null;


	ele = (TextElement)doc.getParent();

	name = ele.getName();
	val  = ele.getTextValue();

	System.err.println("parse_parent: " + name + " - " + val);

	if (name.equals("ESP:REMOTE_INVOKE")) {
	    
	    StructuredTextDocument q;

	    q = _clsExecThread.invoke(doc);
		
	    this.respondQueryOnJXTA2(q);
	    
	} else if (name.equals("ESP:REMOTE_INVOKE_WITH_BYTE")) {

	    System.err.println("enter: ESP REMOTE WITH B");

	    enu = doc.getChildren();
	    
	    ele = (TextElement)enu.nextElement();
	    ele = (TextElement)enu.nextElement();
	    name = ele.getName();
	    val  = ele.getTextValue();

	    int loopcount = Integer.parseInt(val);

	    System.out.println("loopcount: " + loopcount);
	    
	    this._responded = false;
	    
	    StructuredTextDocument q = null;

	    byte[] bArray;
	    
	    _jsReaderThread.suspend();
	    
	    sendDataReadyQuery();

	    _jsReaderThread.suspend();
	    
	    for (int i = 0;i < loopcount; i++) {

		ele = (TextElement)enu.nextElement();
		name = ele.getName();
		val  = ele.getTextValue();

		int size = Integer.parseInt(val);
		DataListenerThread dlt = new DataListenerThread(this, _jxtaSocket, size);
		
		dlt.read();
		bArray = dlt.getByteArray();

		//System.out.println("data: -> " + bArray[0] + " " + bArray[20000]);

		ByteBuffer bbff = ByteBuffer.wrap(bArray);

		_dataVector.add(bbff);
	    }
	    
	    _jsReaderThread.suspend();
	    
	    q = _clsExecThread.invoke(doc);

	    this._responded = true;
	    
	    this.respondQueryOnJXTA2(q);

	    _jsReaderThread.wakeUp();

	    // clean data vector elements
	    _dataVector.removeAllElements();
	    
	} else if (name.equals("ESP:REMOTE_RESPONSE")) {

	    this._returnDoc = doc;
	    this._responded = true;
	    
	} else if (name.equals("ESP:MODULE_SEND")) {
		       
	} else if (name.equals("ESP:MODULE_ACCEPT")) {

	} else if (name.equals("ESP:KEY_SEND")) {

	} else if (name.equals("ESP:KEY_ACCEPT")) {

	} else {

	}	


    }
    
    /************************************************************
     * Fundamental Operantions 
     ************************************************************/
    private boolean initEnv() {
	_shellEnv = getEnv();
	_shellObj = _shellEnv.get("stdgroup");

	_peerGroup = (PeerGroup)_shellObj.getObject();

	_discoveryService = _peerGroup.getDiscoveryService();

	_myName = _peerGroup.getPeerName();

	_dataVector = new Vector();
	
	return true;
    }

    private boolean delete() {

	if (this._jxtaServerSocket != null) {
	    try {
		_jxtaServerSocket.close();
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}

	if (this._jxtaSocket != null) {
	    try {
		_jxtaSocket.close();
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}
	
	return true;
    }

    private void getOpt(String[] args) {
	GetOpt parser = new GetOpt(args, "m:f:c:");
	int option;

	while((option = parser.getNextOption()) != -1) {
	    
	    switch(option) {
		
		// Mode
	    case 'm': {
		_peerMode = Integer.parseInt(parser.getOptionArg());

		if (_peerMode < 0 || 3 < _peerMode) {
		    _peerMode = -1;
		}

		break;
	    }

		// Class Name
	    case 'c': {
		_className = parser.getOptionArg();

		break;
	    }
		
	    case 'f': {
	    }
		break;
	    }

	}
	
    }
    
    /************************************************************
     * Jxta Socket Operations
     ************************************************************/
    public StructuredTextDocument sendQueryOnJXTA2(StructuredTextDocument source) {

	ByteArrayInputStream bais  = null;
	StructuredTextDocument doc = null;

	MessageListenerThread tt =  new MessageListenerThread(this, _jxtaSocket);
	
	String docStringed = null;
	byte[] sourceArray = null;
	
	docStringed = source.toString();
	sourceArray = docStringed.getBytes();

	bais = new ByteArrayInputStream(sourceArray);

	// wait for response
	_jsReaderThread.suspend();
	Thread t = new Thread(tt);
	try {
	    t.start();
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}
	_jsReaderThread.write(bais);
	System.err.println(this.getPeerName() + ": sended query");
	try {
	    t.join();
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}
	//_jsReaderThread.wakeUp();
	
	return tt.getJxtaQuery();
    }

    public void sendQueryOnJXTA2Block(StructuredTextDocument source) {

	ByteArrayInputStream bais  = null;
	StructuredTextDocument doc = null;
	
	String docStringed = null;
	byte[] sourceArray = null;
	
	docStringed = source.toString();
	sourceArray = docStringed.getBytes();

	bais = new ByteArrayInputStream(sourceArray);

	// wait for response
	_jsReaderThread.suspend();
	_jsReaderThread.write(bais);
	
	System.err.println(this.getPeerName() + ": sended query");
    }
    

    public void respondQueryOnJXTA2(StructuredTextDocument source) {

	ByteArrayInputStream bais  = null;
	StructuredTextDocument doc = null;

	String docStringed = null;
	byte[] sourceArray = null;
	
	docStringed = source.toString();
	sourceArray = docStringed.getBytes();

	bais = new ByteArrayInputStream(sourceArray);
	
	_jsReaderThread.write(bais);
    }

    public boolean sendFileOnJXTA2(File file) {

	File f = null;
	FileInputStream fis = null;
	
	try {
	    f   = file;
	    fis = new FileInputStream(f);
	} catch (FileNotFoundException e) {
	    System.err.println(e.getMessage());
	    return false;
	}

	System.err.println(this.getPeerName() + ": file size" + f.length());
	
	_jsReaderThread.write(fis);

	try {
	    fis.close();
	} catch (IOException e) {
	    System.err.println(e.getMessage());
	    return false;
	} finally {
	    fis = null;
	}
	
	return true;
    }

    public boolean sendDataOnJXTA2(byte[] b) {
	ByteArrayInputStream bais  = null;
	
	bais = new ByteArrayInputStream(b);
	
	_jsReaderThread.write(bais);
	
	return true;
    }

    private void sendDataReadyQuery() {
	// Query
	StructuredTextDocument doc
	    = (StructuredTextDocument)StructuredDocumentFactory.newStructuredDocument(new MimeMediaType("text/xml"),
										      "ESP:DATA_READY");
	Element e0 = doc.createElement("size", "10000");
	doc.appendChild(e0);

	// Write to Socket
	ByteArrayInputStream bais  = null;
	
	String docStringed = null;
	byte[] sourceArray = null;
	
	docStringed = doc.toString();
	sourceArray = docStringed.getBytes();

	bais = new ByteArrayInputStream(sourceArray);

	_jsReaderThread.write(bais);
    }

    public byte[] getDataFromJXTA2(int size) {
	DataListenerThread dlt = new DataListenerThread(this, _jxtaSocket, size);

	_jsReaderThread.suspend();
	dlt.read();
	_jsReaderThread.wakeUp();
	    
	return dlt.getByteArray();
    }
    
    public void idleState(int sec) {
	try {
	    Thread.sleep(sec * 1000);
	} catch (Exception e) {
	    ;
	}
    }

    public StructuredTextDocument getResponseDoc() {
	return _returnDoc;
    }

    /************************************************************
     * Used by Working Object
     ************************************************************/
    public void beginInvocation() {

    }

    // get return query from remote peer
    public void doneInvocation() {
	
	MessageListenerThread tt =  new MessageListenerThread(this, _jxtaSocket);
	
	// wait for response
	_jsReaderThread.suspend();
	Thread t = new Thread(tt);
	try {
	    t.start();
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}
	try {
	    t.join();
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}
	_jsReaderThread.wakeUp();
	
	if (this._responded == false) {
	    System.out.println("?????????????????????");
	    this.idleState(5);
	}
    }

    // return only byte array vector
    public Enumeration getMethodArgs() {
	return _dataVector.elements();
    }
    
    /************************************************************
     * Jxta Server Socket Operations
     ************************************************************/
    private boolean createJxtaServerSocket(PipeAdvertisement adv) {
	
	try {
	    _jxtaServerSocket = new JxtaServerSocket();
	    _jxtaServerSocket.bind(_peerGroup, adv);
	} catch (IOException e) {
	    System.err.println(e.getMessage());
	}
	
	return true;
    }

    private boolean acceptJxtaServerSocket(int timeout) {

	try {
	    _jxtaSocket.setSoTimeout(timeout);
	} catch (IOException e) {
	    System.err.println(e.getMessage());
	    return false;
	}
	
	try {
	    _jxtaSocket = (JxtaSocket)_jxtaServerSocket.accept();
	} catch (IOException e) {
	    System.err.println(e.getMessage());
	    return false;
	}

	return true;
    }

    private JxtaSocket connectJxtaSocketByPipeAdvertisement(PipeAdvertisement adv, int timeout) {
	JxtaSocket socket = new JxtaSocket();

	try {
	    socket.connect(_peerGroup, adv, timeout);
	    System.err.println(_myName + ": connect to socket");
	    return socket;
	} catch (IOException e) {
	    _jxtaSocket = null;
	    System.out.println(e.getMessage());
	    return null;
	}
    }
    
    /************************************************************
     * Advertisement Operations
     ************************************************************/
    private boolean createPipeAdvertisement(String name) {
	_pipeAdv = (PipeAdvertisement)AdvertisementFactory.newAdvertisement(PipeAdvertisement.getAdvertisementType());
	
	if (_pipeAdv == null) {
	    System.err.println("set up pipe failes");
	    return false;
	}

	_pipeAdv.setName(name);
	_pipeAdv.setType(PipeService.UnicastType);	
	_pipeAdv.setPipeID((ID) IDFactory.newPipeID(_peerGroup.getPeerGroupID()));

	return true;
    }

    private boolean publishPipeAdvertisementLocal() {

	try {
	    _discoveryService.publish(_pipeAdv);
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	    System.err.println("pipe adv pulish failed");
	    
	    return false;
	}

	return true;	
    }

    private boolean publishPipeAdvertisementRemote() {

	try {
	    _discoveryService.remotePublish(_pipeAdv);
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	    System.err.println("pipe adv pulish failed");

	    return false;
	}

	return true;	
    }

    /*
     * update local advertisement cached list
     */
    private Enumeration getLocalAdvertisementEnum() {

	Enumeration en;
	
	try {
	    en = _discoveryService.getLocalAdvertisements(2,
							  null,
							  null);
	    return en;
	    
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}

	return null;
    }

    /*
     * update remote advertisement cached list
     */
    private int getRemoteAdvertisement(int type) {

	int queryID = 0;
	
	try {
	    queryID = _discoveryService.getRemoteAdvertisements(null,
								type,
								null,
								null,
								5,
								null);
	    return queryID;
	    
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	}

	return -1;
    }

    /*
     * find advertisement
     *
     * not so useful
     */
    private Advertisement findAdvertisement(Enumeration en, String targetType) {

	Advertisement adv = null;

	while (en.hasMoreElements() == true) {
	    adv = (Advertisement)en.nextElement();
	    
	    if ((adv.getAdvType()).equals(targetType)) {
		System.out.println(AdvertisementUtilities.advertisementToXmlText(adv));
		
		return adv;
	    }
	}

	return adv;
    }

    private JxtaSocket findAvailableJxtaSocket(Enumeration en) {
	JxtaSocket socket     = null;
	PipeAdvertisement adv = null;

	while(en.hasMoreElements() == true) {
	    
	    adv = (PipeAdvertisement)findAdvertisement(en,
						       "jxta:PipeAdvertisement");

	    // skip my own pipe advertisement
	    if ((adv.getName()).equals(_myName)) {
		continue;
	    }
	    
	    socket = connectJxtaSocketByPipeAdvertisement(adv, 1 * 1000);

	    if (socket != null) {
		break;
	    }
	}

	return socket;
    }				       
    

    private boolean flushAdvertisement(String targetType) {
	Enumeration en = null;
	Advertisement adv = null;
	
	try {
	    en = _discoveryService.getLocalAdvertisements(2,
							  null,
							  null);
	    
	    while (en.hasMoreElements() == true) {
		adv = (Advertisement)en.nextElement();
		if ((adv.getAdvType()).equals(targetType)) {
		    _discoveryService.flushAdvertisement(adv);
		    System.out.println("jxta: flushed advertisement");
		}
	    }

	} catch (IOException e) {
	    System.out.println(e.getMessage());
	    return false;
	}

	return true;
    }

    private boolean flushPipeAdvertisement() {
	Enumeration en = null;
	Advertisement adv = null;
	PipeAdvertisement pipe = null;
	
	try {
	    en = _discoveryService.getLocalAdvertisements(2,
							  null,
							  null);
	    
	    while (en.hasMoreElements() == true) {
		adv = (Advertisement)en.nextElement();
		
		if ((adv.getAdvType()).equals("jxta:PipeAdvertisement")) {

		    pipe = (PipeAdvertisement)adv;
		    
		    if ((pipe.getName()).equals(_myName)) {
			_discoveryService.flushAdvertisement(pipe);
			System.out.println("jxta: flushed advertisement");
		    }
		}
	    }

	} catch (IOException e) {
	    System.out.println(e.getMessage());
	    return false;
	} 

	return true;
    }
    

    private boolean flushAdvertisement(String targetType, String name) {
	Enumeration en = null;
	Advertisement adv = null;
	
	try {
	    en = _discoveryService.getLocalAdvertisements(2,
							  null,
							  null);
	    
	    while (en.hasMoreElements() == true) {
		adv = (Advertisement)en.nextElement();
		if ((adv.getAdvType()).equals(targetType)) {
		    _discoveryService.flushAdvertisement(adv);
		    System.out.println("jxta: flushed advertisement");
		}
	    }

	} catch (IOException e) {
	    System.out.println(e.getMessage());
	    return false;
	}

	return true;
    }
    
    
    /************************************************************
     * Load Decode Call Refrection
     ************************************************************/
    private Class loadClassByName(String name) {

	Class  cls = null;

	if (name == null) {
	    return null;
	}
	
	try {
	    cls = Class.forName(name);
	} catch (Exception e) {
	    System.err.println(e.getMessage());
	    return null;
	}
	
	return cls;
    }

    private boolean loadEncryptedClass() {
	return true;
    }

    /************************************************************
     * Outlet methods
     ************************************************************/
    public String getPeerName() {
	return this._myName;
    }

    /************************************************************
     * Class Executer Thread
     *
     * Thread
     ************************************************************/
    class ClassExecuterThread implements Runnable {

	private Prototype parent = null;

	private Class  cls = null;
	private Object obj = null;
	private Method met = null;

	private boolean working = false;

	public ClassExecuterThread(Class cls, Prototype parent) {
	    this.cls = cls;
	    this.parent = parent;
	}

	public ClassExecuterThread(Class cls) {
	    this.cls = cls;

	    try {
		obj = cls.newInstance();
	    } catch (InstantiationException e) {
		System.err.println(e.getMessage());
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void init(Prototype parent) {
	    this.parent = parent;

	    Class[]  clss = new Class[1];
	    clss[0] = parent.getClass();
	    
	    Object[] args = new Object[1];
	    args[0] = this.parent;
	    
	    try {
		obj = cls.newInstance();
	    } catch (InstantiationException e) {
		System.err.println(e.getMessage());
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    }
	    
	    try {
		met = cls.getMethod("initWithParam", clss);
	    } catch (NoSuchMethodException e) {
		System.err.println(e.getMessage());
	    }

	    try {
		met.invoke(obj, args);
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    } catch (InvocationTargetException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void run() {
	    
	    Class[]  clss = new Class[1];
	    clss[0] = parent.getClass();

	    Object[] args = new Object[1];
	    args[0] = this.parent;
	    
	    try {
		obj = cls.newInstance();
	    } catch (InstantiationException e) {
		System.err.println(e.getMessage());
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    }
	    
	    try {
		met = cls.getMethod("perform", clss);
	    } catch (NoSuchMethodException e) {
		System.err.println(e.getMessage());
	    }

	    try {
		this.working = true;
		met.invoke(obj, args);
		this.working = false;
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    } catch (InvocationTargetException e) {
		System.err.println(e.getMessage());
	    }

	    System.err.println("exit thread state: " + (this.getClass()).toString());
	}

	public void die() {
	    
	    try {
		met = cls.getMethod("die", null);
	    } catch (NoSuchMethodException e) {
		System.err.println(e.getMessage());
	    }

	    try {
		met.invoke(obj, null);
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    } catch (InvocationTargetException e) {
		System.err.println(e.getMessage());
	    } 
	}

	public StructuredTextDocument invoke(StructuredTextDocument doc) {
	    Class[]  clss = new Class[1];
	    clss[0] = doc.getClass();
	    
	    Object[] args = new Object[1];
	    args[0] = doc;
	    
	    try {
		met = cls.getMethod("invokeMethod", clss);
	    } catch (NoSuchMethodException e) {
		System.err.println(e.getMessage());
		System.out.println("get method");
	    }
	    
	    try {
		doc = (StructuredTextDocument)met.invoke(obj, args);
	    } catch (IllegalAccessException e) {
		System.err.println(e.getMessage());
	    } catch (InvocationTargetException e) {
		System.err.println(e.getMessage());
	    }
	    
	    return doc;
	}
	
    }
    
    /************************************************************
     * Jxta Server Socket Listening Thread
     *
     * Thread
     ************************************************************/    
    class JxtaServerSocketListenerThread implements Runnable {

	private Prototype parent;
	private JxtaServerSocket ssocket;
	private JxtaSocket        socket = null;

	private PeerGroup group;
	private PipeAdvertisement adv;
	
	private boolean reTry   = true;
	private boolean working   = false;
	private boolean connected = false;

	public JxtaServerSocketListenerThread(Prototype parent,
					      JxtaServerSocket ssocket,
					      PeerGroup group,
					      PipeAdvertisement adv) {
	    this.parent  = parent;
	    this.ssocket = ssocket;
	    this.group   = group;
	    this.adv     = adv;

	    this.working   = true;
	    this.connected = false;
	}

	public void run() {

	    int count = 0;
	    this.reTry = true;

	    System.err.println(parent.getPeerName() + ": starting listener thread");	    
	    
	    try {
		ssocket.bind(group, adv);
	    } catch (IOException e) {
		System.err.println(e.getMessage());
		return;
	    }
	    
	    while (this.reTry == true &&
		   this.connected == false &&
		   this.working == true) {
		
		try {
		    socket = (JxtaSocket)ssocket.accept();
		    System.err.println(parent.getPeerName() + ": acceptd.");
		    this.connected = true;
		    break;
		} catch (IOException e) {
		    System.err.println(e.getMessage());
		}
	    }

	    System.err.println("exit thread state: " + (this.getClass()).toString());
	    
	}

	public void die() {
	    this.working = false;
	}

	public JxtaSocket getSocket() {
	    return this.socket;
	}

	public boolean isConnected() {
	    return this.connected;
	}
    }

    /************************************************************
     * JxtaSocket Read Thread
     *
     * Thread
     ************************************************************/
    class JxtaSocketReaderThread implements Runnable {

	private Prototype parent = null;

	public boolean alive;
	public boolean suspend;

	private JxtaSocket   socket = null;
	private InputStream  input  = null;
	private OutputStream output = null;

	private byte[] buf = null;

	private StructuredTextDocument currentDoc;

	public JxtaSocketReaderThread(Prototype parent, JxtaSocket socket) {
	    this.parent = parent;
	    this.socket = socket;
	    this.alive  = true;
	    this.suspend = false;

	    this.buf = new byte[8192];

	    try {
		
		input  = socket.getInputStream();
		output = socket.getOutputStream();
		
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void run() {
	    long total = 0;
	    int  read = 0;

	    System.err.println(parent.getPeerName() + ": starting reader thread");
	    
	    while(alive == true && read != -1) {
		try {
		    if (suspend == true) {
			continue;
		    }
		    
		    read = input.read(buf);
		    total += read;
		    
		    System.out.println(parent.getPeerName() + ": total " + total + "bytes");
		    
		    System.err.println("handle xml: " + (this.getClass()).toString());
		    parent.handleXMLDocument(buf);

		    if (read > 1) {
			//System.out.println(parent.getPeerName() + ": Read Document ");
			//System.out.println(((StructuredTextDocument)DocumentUtilities.getStructuredXmlDocument(buf)).toString());
		    }
		} catch (Exception e) {
		    System.out.println(e.getMessage());
		}
	    }

	    System.err.println("exit thread state: " + (this.getClass()).toString());
	}

	public void write(byte[] data, int bytes) {
	    try {
		output.write(data);
		output.flush();
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	    
	}

	public void write(InputStream in) {
	    int read;
	    byte[] tmp = new byte[8192];
	    
	    try {
		
		while((read = in.read(tmp)) != -1) {
		    try {
			output.write(tmp, 0, read);
			output.flush();
		    } catch (IOException e) {
			System.err.println(e.getMessage());
		    }
		}

		// flush remain data
		try {
		    output.flush();
		} catch (IOException e) {
		    System.err.println(e.getMessage());
		}
		
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void suspend() {
	    suspend = true;
	}

	public void wakeUp() {
	    System.out.println("wake up");
	    suspend = false;
	}

	public void die() {
	    this.alive = false;
	}
	
	public boolean closeSocket() {
	    try {
		this.socket.close();
	    } catch (Exception e) {
		System.err.println(e.getMessage());
		return false;
	    }

	    return true;
	}
	
	public StructuredTextDocument getJxtaQuery() {
	    StructuredTextDocument doc = null;
	    try {
		doc = (StructuredTextDocument)DocumentUtilities.getStructuredXmlDocument(buf);
	    } catch (JxtaException e) {
		System.out.println(e.getMessage());
	    }

	    return doc;
	}
    }

    /************************************************************
     * JxtaSocket Read Thread
     *
     * Thread
     ************************************************************/
    class MessageListenerThread implements Runnable {

	private Prototype parent = null;


	protected byte[] buf;
	public boolean alive;
	protected int size;

	private JxtaSocket   socket = null;
	private InputStream  input  = null;

	private StructuredTextDocument currentDoc;

	public MessageListenerThread(Prototype parent, JxtaSocket socket) {
	    this.parent = parent;
	    this.socket = socket;
	    this.alive = true;


	    this.buf = new byte[8192];
	    this.size = 0;
	    
	    try {
		input  = socket.getInputStream();
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void run() {

	    long total = 0;
	    int  read = 0;

	    System.err.println(parent.getPeerName() + ": starting response thread");
	    
	    while(read == 0) {
		try {
		    read = input.read(buf);
		    total += read;
		    
		    System.out.println(parent.getPeerName() + ": revcieve response");
		} catch (Exception e) {
		    System.out.println(e.getMessage());
		}
	    }
	    
	    System.err.println("handle xml: " + (this.getClass()).toString());
	    parent.handleXMLDocument(buf);	    

	    System.err.println("exit thread state: " + (this.getClass()).toString());
	}

	public void setSize(int size) {
	    this.size = size;
	}

	public void die() {
	    this.alive = false;
	}
	
	public StructuredTextDocument getJxtaQuery() {
	    
	    StructuredTextDocument doc = null;
	    try {
		doc = (StructuredTextDocument)DocumentUtilities.getStructuredXmlDocument(buf);
	    } catch (JxtaException e) {
		System.out.println(e.getMessage());
	    }

	    return doc;
	}
    }

    /************************************************************
     * Data Read Thread
     *
     * Thread
     ************************************************************/
    class DataListenerThread  implements Runnable {
	private Prototype parent = null;
	private ByteArrayOutputStream baos = null;

	protected byte[] buf;
	public boolean alive;
	protected int size;
	
	private JxtaSocket   socket = null;
	private InputStream  input  = null;
	
	private StructuredTextDocument currentDoc;
	
	public DataListenerThread(Prototype parent, JxtaSocket socket, int size) {
	    this.parent = parent;
	    this.socket = socket;
	    this.alive = true;
	    this.size = size;
	    this.buf = new byte[size];
	    this.baos = new ByteArrayOutputStream(size);
	    
	    try {
		input  = socket.getInputStream();
	    } catch (IOException e) {
		System.err.println(e.getMessage());
	    }
	}

	public void run() {
	    ;
	}

	public byte[] read() {

	    long total = 0;
	    int  read = 0;
	    
	    while(total < size) {
		try {
		    read = input.read(buf, 0, size);
		    
		    baos.write(buf, 0, read);
		    
		    total += read;
		    //System.err.println("total : " + total + "bytes");
		} catch (Exception e) {
		    System.out.println(e.getMessage());
		}
	    }

	    System.err.println("read : " + total + "bytes");
	    return buf;
	}

	public byte[] getByteArray() {
	    return baos.toByteArray();
	}
	
    }

    
}
